import os
import sys
import math
from rank_bm25 import BM25Okapi

import Orange
from Orange.widgets import widget
from Orange.widgets.utils.signals import Input, Output
from Orange.data import  ContinuousVariable
from thefuzz import fuzz
from AnyQt.QtWidgets import QApplication, QComboBox
from Orange.widgets.settings import Setting

# Import intelligent selon contexte
if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils import thread_management
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils import thread_management
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file


@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWKeywordsDetection(widget.OWWidget):
    name = "Keywords Detection"
    description = 'The input Data must contain a column "content". The input Keywords must contain a column "keywords". This widget will count the number of keywords that occur in the content with a fuzzy matching (percentage based on small variations).'
    category = "AAIT - LLM INTEGRATION"
    icon = "icons/owkeywordsdetection.png"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owkeywordsdetection.png"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owkeywordsdetection.ui")
    want_control_area = False
    priority = 1071

    # Settings
    mode = Setting("Fuzzy")

    class Inputs:
        data = Input("Data", Orange.data.Table)
        keywords = Input("Keywords", Orange.data.Table)

    class Outputs:
        data = Output("Data", Orange.data.Table)


    @Inputs.data
    def set_data(self, in_data):
        self.data = in_data
        if self.autorun:
            self.run()

    @Inputs.keywords
    def set_keywords(self, in_keywords):
        self.keywords = in_keywords
        if self.autorun:
            self.run()

    def __init__(self):
        super().__init__()
        # Qt Management
        self.setFixedWidth(470)
        self.setFixedHeight(300)
        uic.loadUi(self.gui, self)

        self.combobox_mode = self.findChild(QComboBox, "comboBox")
        self.combobox_mode.setCurrentIndex(self.combobox_mode.findText(self.mode))
        self.combobox_mode.currentTextChanged.connect(self.on_mode_changed)

        # Data Management
        self.data = None
        self.keywords = None
        self.thread = None
        self.autorun = True
        self.result = None

        # Custom updates
        self.post_initialized()

    def on_mode_changed(self, text):
        self.mode = text
        self.run()


    def run(self):
        self.error("")
        self.warning("")

        # If Thread is already running, interrupt it
        if self.thread is not None:
            if self.thread.isRunning():
                self.thread.safe_quit()

        if self.data is None:
            self.Outputs.data.send(None)
            return

        if self.keywords is None:
            self.Outputs.data.send(None)
            return

        if "content" not in self.data.domain:
            self.error('You need a "content" column in your input Data.')
            self.Outputs.data.send(None)
            return

        if "keywords" not in self.keywords.domain:
            self.error('You need a "keywords" column in your input Keywords.')
            self.Outputs.data.send(None)
            return


        self.progressBarInit()

        # Connect and start thread : main function, progress, result and finish
        # --> progress is used in the main function to track progress (with a callback)
        # --> result is used to collect the result from main function
        # --> finish is just an empty signal to indicate that the thread is finished
        self.thread = thread_management.Thread(compute_keywords_on_table, self.data, self.keywords, self.mode)
        self.thread.progress.connect(self.handle_progress)
        self.thread.result.connect(self.handle_result)
        self.thread.finish.connect(self.handle_finish)
        self.thread.start()

    def handle_progress(self, value: float) -> None:
        self.progressBarSet(value)

    def handle_result(self, result):
        try:
            self.result = result
            self.Outputs.data.send(result)
        except Exception as e:
            print("An error occurred when sending out_data:", e)
            self.Outputs.data.send(None)
            return

    def handle_finish(self):
        print("Keywords computation finished !")
        self.progressBarFinished()

    def post_initialized(self):
        pass



def compute_keywords_on_table(table, keywords, mode="Fuzzy", progress_callback=None, argself=None):
    """
    Compute fuzzy match scores between table rows and a list of keywords,
    and add them as a new meta column.

    Args:
        table (Orange.data.Table): Input table with a "content" column.
        keywords (Orange.data.Table): Table with a "keywords" column.
        mode (str): Type of keyword search, "Fuzzy" or "Strict".
        progress_callback (callable, optional): Function called with progress (0–100).
        argself (object, optional): If has attribute `stop=True`, stops processing early.

    Returns:
        Orange.data.Table: Copy of input with a new meta column "Score - Keywords".
    """
    # Make a copy of the table to avoid modifying the original
    data = table.copy()

    # Extract the list of keywords from the 'keywords' table
    keywords_list = [row["keywords"].value for row in keywords]

    if mode == "Elastic":
        corpus = [row["content"].value for row in data]
        scores = get_bm25_scores(corpus, keywords_list)

    else:
        scores = []
        # Iterate over each row in the copied table
        for i, row in enumerate(data):
            content = row["content"].value  # Text from the 'content' column
            score = 0
            if mode == "Fuzzy":
                score = fuzzy_match_score(content, keywords_list)  # Compute similarity score
            elif mode == "Strict":
                score = strict_match_score(content, keywords_list)

            scores.append(score)

            # Update progress if a progress callback is provided
            if progress_callback is not None:
                progress_value = float(100 * (i + 1) / len(data))
                progress_callback(progress_value)

            # Stop loop if 'argself' exists and stop flag is set
            if argself is not None:
                if argself.stop:
                    break

    # Create a new continuous variable (column) for keyword scores
    score_var = ContinuousVariable("Score - Keywords")
    # Add the scores as a meta attribute to the table
    data = data.add_column(score_var, scores, to_metas=True)

    return data



def fuzzy_match_score(text, keywords_list):
    """
    Checks if keywords are present in text using fuzzy matching
    and returns a global score.

    Args:
        text (str): The full text to search in.
        keywords_list (list): List of keywords to find.

    Returns:
        float: Global match score (0-100).
    """
    keywords_list = list(filter(None, keywords_list))
    if len(keywords_list)==0:
        return 0

    words = text.split(" ")  # Split text into words
    total_score = 0

    for keyword in keywords_list:
        best_score = max(fuzz.ratio(word.lower(), keyword.lower()) for word in words)
        total_score += best_score

    return total_score / len(keywords_list)  # Normalize score


def strict_match_score(text, keywords_list):
    """
    Checks if keywords are present in text using exact matching
    and returns a global score (0 or 100).

    Args:
        text (str): The full text to search in.
        keywords_list (list): List of keywords to find.

    Returns:
        float: Global match score (0-100).
    """
    keywords_list = list(filter(None, keywords_list))
    if len(keywords_list) == 0:
        return 0.0

    # Normalize text and split into words
    words = set(text.lower().split())

    matches = sum(1 for keyword in keywords_list if keyword.lower() in words)
    score = (matches / len(keywords_list)) * 100
    return score


### Gemini freestyle
def bm25_match_score(text, keywords_list, idf_map=None, k1=1.5, b=0.75, avg_dl=10.0):
    """
    Calculates the BM25 score for a text given a list of keywords.

    Args:
        text (str): The document text.
        keywords_list (list): The query terms.
        idf_map (dict): Precomputed IDF scores for terms.
                        If None, defaults to 1.0 for all terms.
        k1 (float): Term frequency saturation parameter (usually 1.2 to 2.0).
        b (float): Length normalization parameter (usually 0.75).
        avg_dl (float): Average length of documents in the collection.

    Returns:
        float: The BM25 relevance score.
    """
    if not keywords_list or not text:
        return 0.0

    # Tokenize and normalize
    doc_words = text.lower().split()
    doc_len = len(doc_words)

    # Calculate term frequencies in this specific document
    tf_counts = {}
    for word in doc_words:
        tf_counts[word] = tf_counts.get(word, 0) + 1

    score = 0.0
    for keyword in keywords_list:
        keyword = keyword.lower()
        if keyword not in tf_counts:
            continue

        # Get IDF (default to 1.0 if not provided)
        idf = idf_map.get(keyword, 1.0) if idf_map else 1.0

        # Calculate TF component with saturation and length normalization
        tf = tf_counts[keyword]
        numerator = tf * (k1 + 1)
        denominator = tf + k1 * (1 - b + b * (doc_len / avg_dl))

        score += idf * (numerator / denominator)

    return score

### Gemini freestyle
def calculate_idf(corpus):
    """
    Args:
        corpus (list of str): All documents in your dataset.
    """
    N = len(corpus)
    idf_map = {}
    words_in_docs = [set(doc.lower().split()) for doc in corpus]

    # Find all unique words across all docs
    all_unique_words = set().union(*words_in_docs)

    for word in all_unique_words:
        # Number of documents containing the word
        n_q = sum(1 for doc_set in words_in_docs if word in doc_set)
        # Standard BM25 IDF formula
        idf_map[word] = math.log((N - n_q + 0.5) / (n_q + 0.5) + 1.0)

    return idf_map



def get_bm25_scores(corpus, keywords_list):
    """
    Calcule les scores BM25 pour un corpus entier par rapport à des mots-clés.

    Args:
        corpus (list of str): Liste de textes (documents).
        keywords_list (list of str): Liste des mots-clés de recherche.

    Returns:
        list of float: Les scores pour chaque document du corpus.
    """
    # 1. Préparation des données (Tokenization simple)
    # La bibliothèque attend une liste de listes de mots
    tokenized_corpus = [doc.lower().split() for doc in corpus]
    tokenized_query = [keyword.lower() for keyword in keywords_list]

    # 2. Initialisation du modèle BM25
    # C'est ici que k1, b et avg_dl sont calculés/appliqués
    bm25 = BM25Okapi(tokenized_corpus)

    # 3. Calcul des scores
    scores = bm25.get_scores(tokenized_query)

    return scores



if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWKeywordsDetection()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()




import os
from AnyQt.QtWidgets import QApplication, QLineEdit
import Orange.data
from Orange.widgets import widget
from Orange.widgets.utils.signals import Input, Output
from Orange.widgets.settings import Setting

if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file

@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWExtraChunks(widget.OWWidget):
    name = "Extra Chunks"
    description = "Extract surrounding chunks from a dataset"
    category = "AAIT - LLM INTEGRATION"
    icon = "icons/extra_chunks.png"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/extra_chunks.png"
    want_control_area = False
    priority = 1001
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owextrachunks.ui")

    class Inputs:
        complete_data = Input("Complete Dataset", Orange.data.Table)
        selected_data = Input("Chunks", Orange.data.Table)

    class Outputs:
        data = Output("Data", Orange.data.Table)
        concatenate_data = Output("Concatenate Data", Orange.data.Table)

    extra_chunks: int = Setting(2)

    @Inputs.complete_data
    def set_complete_data(self, data):
        self.complete_data = data
        if self.autorun:
            self.process()

    @Inputs.selected_data
    def set_selected_data(self, data):
        self.selected_data = data
        if self.autorun:
            self.process()

    def __init__(self):
        super().__init__()
        self.setFixedWidth(470)
        self.setFixedHeight(300)
        uic.loadUi(self.gui, self)

        self.complete_data = None
        self.selected_data = None
        self.autorun = True

        self.edit_extrachunks = self.findChild(QLineEdit, 'QExtraChunks')
        self.edit_extrachunks.setText(str(self.extra_chunks))
        self.edit_extrachunks.textChanged.connect(self.update_extrachunks)

    def update_extrachunks(self, text):
        self.extra_chunks = int(text) if text.isdigit() else 2

    def process(self):
        self.error("")
        self.warning("")

        if self.complete_data is None or self.selected_data is None:
            self.Outputs.concatenate_data.send(None)
            self.Outputs.data.send(None)
            return

        domain = self.complete_data.domain
        if "path" not in domain:
            self.error('You need a "path" column in your input tables.')
            self.Outputs.concatenate_data.send(None)
            self.Outputs.data.send(None)
            return

        if "Chunks index" not in domain:
            self.error('You need a "Chunks index" column in your input tables.')
            self.Outputs.concatenate_data.send(None)
            self.Outputs.data.send(None)
            return

        if "Chunks" not in domain:
            self.error('You need a "Chunks" column in your input tables.')
            self.Outputs.concatenate_data.send(None)
            self.Outputs.data.send(None)
            return

        index_var = domain["Chunks index"]
        path_var = domain["path"]
        chunks_var = domain["Chunks"]

        selected_indices_by_path = {}
        for row in self.selected_data:
            path_value = row[path_var]
            index_value = int(row[index_var])
            if path_value not in selected_indices_by_path:
                selected_indices_by_path[path_value] = []
            selected_indices_by_path[path_value].append(index_value)

        complete_indices_by_path = {}
        complete_chunks_by_path = {}
        for row in self.complete_data:
            path_value = row[path_var]
            index_value = int(row[index_var])
            text_chunk = row[chunks_var]
            if path_value not in complete_indices_by_path:
                complete_indices_by_path[path_value] = []
                complete_chunks_by_path[path_value] = {}
            complete_indices_by_path[path_value].append(index_value)
            complete_chunks_by_path[path_value][index_value] = text_chunk

        full_indices = set()
        for path_value, selected_indices in selected_indices_by_path.items():
            if path_value in complete_indices_by_path:
                complete_indices = complete_indices_by_path[path_value]
                min_idx, max_idx = min(complete_indices), max(complete_indices)
                for idx in selected_indices:
                    start_idx = max(min_idx, idx - self.extra_chunks)
                    end_idx = min(max_idx, idx + self.extra_chunks)
                    full_indices.update((path_value, i) for i in range(start_idx, end_idx + 1))

        selected_rows = [row for row in self.complete_data if (row[path_var], int(row[index_var])) in full_indices]
        output_data = Orange.data.Table(self.complete_data.domain, selected_rows)
        self.Outputs.data.send(output_data)

        merged_results = self.merge_chunks(selected_rows)
        concat_domain = Orange.data.Domain([], metas=[
            Orange.data.StringVariable("path"),
            Orange.data.StringVariable("Merged Chunks")
        ])
        merged_table = Orange.data.Table.from_list(concat_domain, merged_results)
        self.Outputs.concatenate_data.send(merged_table)

    def remove_overlap(self, text1, text2):
        if not text1 or text1 == Orange.data.Value:
            return str(text2) if text2 else ""
        if not text2 or text2 == Orange.data.Value:
            return str(text1) if text1 else ""

        text1 = str(text1)
        text2 = str(text2)

        max_overlap = min(len(text1), len(text2))
        for i in range(max_overlap, 0, -1):
            if text1[-i:] == text2[:i]:
                return text1 + text2[i:]
        return text1 + text2

    def merge_chunks(self, rows):
        doc_groups = {}
        for row in rows:
            path = row[self.complete_data.domain["path"]]
            chunk = row[self.complete_data.domain["Chunks"]]
            index = int(row[self.complete_data.domain["Chunks index"]])
            if path not in doc_groups:
                doc_groups[path] = {}
            doc_groups[path][index] = chunk  # Stocke les chunks avec leur index

        merged_results = []

        # 2. Parcourir les indices sélectionnés et récupérer leurs voisins
        for row in self.selected_data:
            path = row[self.complete_data.domain["path"]]
            index = int(row[self.complete_data.domain["Chunks index"]])

            if path not in doc_groups:
                continue  # Si le path n'existe pas, on passe

            # Définir les indices voisins à récupérer
            start_idx = index - self.extra_chunks
            end_idx = index + self.extra_chunks

            # Fusionner les chunks environnants
            merged_text = ""
            for i in range(start_idx, end_idx + 1):
                if i in doc_groups[path]:  # Vérifie que l'index existe
                    merged_text = self.remove_overlap(merged_text, doc_groups[path][i])

            merged_results.append([path, merged_text])

        return merged_results


if __name__ == "__main__":
    import sys
    app = QApplication(sys.argv)
    window = OWExtraChunks()
    window.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()


import os
import sys

import Orange.data
from AnyQt.QtWidgets import QApplication, QLabel
from Orange.widgets import widget
from Orange.widgets.utils.signals import Input, Output
from Orange.widgets.settings import Setting
from AnyQt.QtWidgets import QLineEdit, QTextBrowser, QSpinBox, QDoubleSpinBox

if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.llm import answers
    from Orange.widgets.orangecontrib.AAIT.utils import thread_management,MetManagement
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.llm import answers
    from orangecontrib.AAIT.utils import thread_management,MetManagement
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file

#TODO Ajouter un bouton d'autorun

@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWQueryLLM(widget.OWWidget):
    name = "Query LLM"
    description = "Generate a response to a column 'prompt' with a LLM"
    category = "AAIT - LLM INTEGRATION"
    icon = "icons/owqueryllm.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owqueryllm.svg"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owqueryllm.ui")
    want_control_area = True
    priority = 1089

    class Inputs:
        data = Input("Data", Orange.data.Table)
        model_path = Input("Model", str, auto_summary=False)

    class Outputs:
        data = Output("Data", Orange.data.Table)

    n_ctx: str = Setting("32768")
    workflow_id = Setting("")

    max_tokens = Setting(4096)
    temperature = Setting(0.4)
    top_p = Setting(0.4)
    top_k = Setting(40)
    repeat_penalty = Setting(1.15)

    @Inputs.data
    def set_data(self, in_data):
        self.data = in_data
        if self.autorun:
            self.run()

    @Inputs.model_path
    def set_model_path(self, in_model_path):
        self.model_path = in_model_path
        if self.autorun:
            self.run()

    def __init__(self):
        super().__init__()
        # Qt Management
        self.setFixedWidth(700)
        self.setFixedHeight(700)
        uic.loadUi(self.gui, self)
        self.label_description = self.findChild(QLabel, 'Description')
        # Context size
        self.edit_nCtx = self.findChild(QLineEdit, 'lineEdit')
        self.edit_nCtx.setText(str(self.n_ctx))
        self.edit_nCtx.editingFinished.connect(self.update_n_ctx)
        # Workflow ID for streaming
        self.edit_ID = self.findChild(QLineEdit, 'lineEdit_2')
        self.edit_ID.setText(self.workflow_id)
        self.edit_ID.editingFinished.connect(self.update_workflow_id)
        # Bind all spinboxes
        self.edit_max_tokens = self.bind_spinbox("boxMaxTokens", self.max_tokens)
        self.edit_temperature = self.bind_spinbox("boxTemperature", self.temperature, is_double=True)
        self.edit_top_p = self.bind_spinbox("boxTopP", self.top_p, is_double=True)
        self.edit_top_k = self.bind_spinbox("boxTopK", self.top_k)
        self.edit_repeat_penalty = self.bind_spinbox("boxRepeatPenalty", self.repeat_penalty, is_double=True)

        # Text browser
        self.textBrowser = self.findChild(QTextBrowser, 'textBrowser')

        # # Data Management
        self.data = None
        self.model_path = None
        self.thread = None
        self.autorun = True
        self.use_gpu = False
        self.can_run = True
        self.result = None
        self.n_ctx = self.edit_nCtx.text() if self.edit_nCtx.text().isdigit() else "4096"
        self.workflow_id = self.edit_ID.text()

        # Custom updates
        self.post_initialized()

    def update_n_ctx(self):
        value = self.edit_nCtx.text()  # Read the current value
        self.n_ctx = value if value.isdigit() else "4096"  # Default or parsed value
        self.edit_nCtx.setText(self.n_ctx)
        answers.check_gpu(self.model_path, self)
        #self.run()

    def update_workflow_id(self):
        self.workflow_id = self.edit_ID.text()
        #self.run()

    def bind_spinbox(self, name, value, is_double=False):
        widget_type = QDoubleSpinBox if is_double else QSpinBox
        box = self.findChild(widget_type, name)
        box.setValue(value)
        box.editingFinished.connect(self.update_parameters)
        return box

    def update_parameters(self):
        self.max_tokens = self.edit_max_tokens.value()
        self.temperature = self.edit_temperature.value()
        self.top_p = self.edit_top_p.value()
        self.top_k = self.edit_top_k.value()
        self.repeat_penalty = self.edit_repeat_penalty.value()
        print("Updated parameters:")
        print(f"  Max tokens     : {self.max_tokens}")
        print(f"  Temperature    : {self.temperature}")
        print(f"  Top P          : {self.top_p}")
        print(f"  Top K          : {self.top_k}")
        print(f"  Repeat penalty : {self.repeat_penalty}")
        #self.run()


    def run(self):
        # If Thread is already running, interrupt it
        if self.thread is not None:
            if self.thread.isRunning():
                self.thread.safe_quit()

        if self.data is None:
            self.Outputs.data.send(None)
            return

        if self.model_path is None or not os.path.exists(self.model_path):
            self.Outputs.data.send(None)
            return

        if not "prompt" in self.data.domain:
            self.Outputs.data.send(None)
            return

        answers.check_gpu(self.model_path, self)
        if not self.can_run:
            self.Outputs.data.send(None)
            return

        query_parameters = {
            "max_tokens": self.max_tokens,
            "temperature": self.temperature,
            "top_p": self.top_p,
            "top_k": self.top_k,
            "repeat_penalty": self.repeat_penalty
        }
        chemin_dossier = MetManagement.get_api_local_folder(workflow_id=self.workflow_id)
        if os.path.exists(chemin_dossier):
            MetManagement.write_file_time(chemin_dossier + "time.txt")
        # Start progress bar
        self.progressBarInit()
        self.textBrowser.setText("")

        # Connect and start thread : main function, progress, result and finish
        # --> progress is used in the main function to track progress (with a callback)
        # --> result is used to collect the result from main function
        # --> finish is just an empty signal to indicate that the thread is finished
        self.thread = thread_management.Thread(answers.generate_answers, self.data, self.model_path,
                                               self.use_gpu, int(self.n_ctx), query_parameters, self.workflow_id)
        self.thread.progress.connect(self.handle_progress)
        self.thread.result.connect(self.handle_result)
        self.thread.finish.connect(self.handle_finish)
        self.thread.start()

    def handle_progress(self, progress) -> None:
        value = progress[0]
        text = progress[1]
        if value is not None:
            self.progressBarSet(value)
        if text is None:
            self.textBrowser.setText("")
        else:
            self.textBrowser.insertPlainText(text)

    def handle_result(self, result):
        try:
            self.result = result
            self.Outputs.data.send(result)
        except Exception as e:
            print("An error occurred when sending out_data:", e)
            self.Outputs.data.send(None)
            return

    def handle_finish(self):
        print("Generation finished")
        self.progressBarFinished()

    def post_initialized(self):
        pass

if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWQueryLLM()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()



import os
import sys
import numpy as np

import Orange.data
from AnyQt.QtWidgets import QApplication
from Orange.widgets import widget
from Orange.widgets.utils.signals import Input, Output

if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file


@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWSelectRowsDynamic(widget.OWWidget):
    name = "Select Rows Dynamic"
    description = "Select a row from a second entry"
    category = "AAIT - TOOLBOX"
    icon = "icons/select_dynamic_row.png"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/select_dynamic_row.png"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owselect_row_dynamic.ui")
    want_control_area = False
    priority = 1060

    class Inputs:
        data = Input("data", Orange.data.Table)
        data_for_filter = Input("input_for_filtering", Orange.data.Table)

    class Outputs:
        data_matching = Output("Matching Data", Orange.data.Table)
        data_unmatching = Output("UnMatching Data", Orange.data.Table)

    @Inputs.data
    def set_data(self, data_in):
        if data_in is None:
            return
        self.in_data = data_in
        if self.data_filter_in is None:
            return
        self.run()

    @Inputs.data_for_filter
    def set_path_table(self, in_data_filter):
        if in_data_filter is None:
            return

        total_columns = len(in_data_filter.domain.attributes) + len(in_data_filter.domain.class_vars) + len(
            in_data_filter.domain.metas)
        self.error("")
        if total_columns != 1:
            self.error("error filter_input can only use 1 column in this version")
            return
        if len(in_data_filter.domain.metas) != 1:
            self.error("error filter_input can only use Stringvariable")
            return
        if not isinstance(in_data_filter.domain.metas[0], Orange.data.StringVariable):
            self.error("error filter_input can only use Stringvariable.")
            return

        self.data_filter_in = in_data_filter
        if self.in_data is not None:
            self.run()

            # if total_columns != 1:
            #
            #     return
            #
            # print("in_data_filter")
            # print(in_data_filter)

    def __init__(self):
        super().__init__()
        # Qt Management
        self.setFixedWidth(470)
        self.setFixedHeight(300)
        uic.loadUi(self.gui, self)
        self.data_filter_in = None
        self.in_data = None
        self.autorun = True
        self.post_initialized()

    def run(self):
        self.error("")
        filter_var = self.data_filter_in.domain.metas[0]
        filter_var_name = filter_var.name

        # On utilise .index() ou une recherche directe dans le domaine
        match_var = self.in_data.domain[filter_var_name]
        if match_var is None or not match_var.is_string:
            self.error(f"La colonne '{filter_var_name}' est absente ou n'est pas une StringVariable.")
            return

        # On accède directement à l'attribut .metas (Tableau numpy d'objets pour les strings)
        filter_col_idx = self.data_filter_in.domain.metas.index(filter_var)
        filter_values = self.data_filter_in.metas[:, filter_col_idx]
        values_filter_set = {str(v) for v in filter_values if v and v != ""}

        # Au lieu de boucler sur les 'rows', on récupère toute la colonne d'un coup
        data_col_idx = self.in_data.domain.metas.index(match_var)
        data_values = self.in_data.metas[:, data_col_idx]

        # La compréhension sur un array numpy est souvent plus rapide que sur des objets 'Row'
        mask = np.array([str(val) in values_filter_set for val in data_values])

        matched_table = self.in_data[mask] if np.any(mask) else None
        unmatched_table = self.in_data[~mask] if np.any(~mask) else None

        self.Outputs.data_matching.send(matched_table)
        self.Outputs.data_unmatching.send(unmatched_table)


    def post_initialized(self):
        pass

if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWSelectRowsDynamic()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()



import os
import sys

from AnyQt.QtWidgets import QApplication
from Orange.widgets import widget
from Orange.widgets.utils.signals import Output
from transformers import AutoTokenizer, MarianMTModel


if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils import SimpleDialogQt, thread_management
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils.MetManagement import GetFromRemote, get_local_store_path
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.utils import SimpleDialogQt, thread_management
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils.MetManagement import GetFromRemote, get_local_store_path
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file

@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWModel_HelsinkiEnFr(widget.OWWidget):
    name = "Model - Translation - Helsinki EN-FR"
    description = "Load the translation model Helsinki EN-FR from the AAIT Store"
    category = "AAIT - MODELS"
    icon = "icons/owmodel_helsinki_en_fr.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owmodel_helsinki_en_fr.svg"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owmodel_helsinki_en_fr.ui")
    priority = 1101
    want_control_area = False

    class Outputs:
        out_models = Output("Models", (MarianMTModel, object), auto_summary=False)

    def __init__(self):
        super().__init__()
        # Path management
        self.current_ows = ""
        local_store_path = get_local_store_path()
        model_name = "helsinki_en_fr"
        self.model_path = os.path.join(local_store_path, "Models", "NLP", model_name)
        self.model = None
        self.tokenizer = None
        self.models = None
        # Qt Management
        self.setFixedWidth(470)
        self.setFixedHeight(300)
        uic.loadUi(self.gui, self)
        if not os.path.exists(self.model_path):
            if not SimpleDialogQt.BoxYesNo("Model isn't in your computer. Do you want to download it from AAIT store?"):
                return
            try:
                GetFromRemote("Translation")
            except Exception as e:  # TODO ciblage de l'erreur
                SimpleDialogQt.BoxError(f"Unable to get the model: {e}")
                return
        # Data Management
        self.progressBarInit()
        self.thread = thread_management.Thread(self.load_model, self.model_path)
        self.thread.finish.connect(self.handle_loading_finish)
        self.thread.start()

    def load_model(self, model_path):
        self.model = MarianMTModel.from_pretrained(model_path)
        self.tokenizer = AutoTokenizer.from_pretrained(model_path)

    def handle_loading_finish(self):
        self.models = (self.model, self.tokenizer)
        if self.model is not None and self.tokenizer is not None:
            self.Outputs.out_models.send(self.models)
        else:
            SimpleDialogQt.BoxError("An Error Occurred when loading model.")
            self.Outputs.out_models.send(None)
        self.progressBarFinished()


if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWModel_HelsinkiEnFr()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()



import os
import sys

import Orange.data
from AnyQt.QtWidgets import QApplication
from Orange.widgets import widget
from Orange.widgets.utils.signals import Input, Output


if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file

@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWTrigger(widget.OWWidget):
    name = "Trigger"
    description = "Pause a data signal until a trigger signal is received. Be careful to connect the trigger signal AFTER the data signal."
    category = "AAIT - TOOLBOX"
    icon = "icons/owtrigger.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owtrigger.svg"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owtrigger.ui")
    want_control_area = False
    priority = 1060

    class Inputs:
        data = Input("Data", Orange.data.Table)
        trigger = Input("Trigger", object, auto_summary=False)

    class Outputs:
        data = Output("Data", Orange.data.Table)

    @Inputs.data
    def set_data(self, in_data):
        self.data = in_data
        self.Outputs.data.send(None)

    @Inputs.trigger
    def set_trigger(self, in_trigger):
        self.trigger = in_trigger
        self.run()

    def __init__(self):
        super().__init__()
        # Qt Management
        self.setFixedWidth(470)
        self.setFixedHeight(300)
        uic.loadUi(self.gui, self)

        # Data Management
        self.data = None
        self.trigger = None
        self.autorun = True
        self.post_initialized()

    def run(self):
        if self.trigger is None:
            return
        if self.data is not None:
            self.Outputs.data.send(self.data)

    def post_initialized(self):
        pass

if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWTrigger()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()



import os
from AnyQt.QtWidgets import QWidget, QVBoxLayout, QLineEdit, QListWidget, QListWidgetItem
from AnyQt.QtCore import pyqtSignal
import Orange
from Orange.widgets import gui, widget
from Orange.widgets.settings import Setting
from Orange.widgets.utils.signals import Input, Output

if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
else:
    from orangecontrib.AAIT.utils.import_uic import uic


class BaseListWidget(widget.OWWidget, openclass=True):
    """
    Base Orange widget providing a filterable list of variables from the input data domain.

    This class automatically:
      - Adds a `FilterableVariableList` UI component.
      - Handles input data and populates the list with variables.
      - Restores the last selected variable (`selected_column_name`).

    ______________________________________________________________________________________
    How to use:
    ⚠️ The linked .ui file **must include a `QWidget` named "placeholder"`**, where the
    variable list will be inserted (see owexecutescript_TEST.ui).

    1. Inherit from `BaseListWidget` instead of `Orange.widgets.OWWidget`.
       Your widget will automatically have the attribute `selected_column_name`.

    2. Use `selected_column_name` in your `run()` function to determine
       which column should be processed. For example, check if it's in
       `data.domain` or verify its type (e.g., `StringVariable`).

    3. Pass `selected_column_name` to any threaded function you use:
       ```python
       thread_management.Thread(self.my_function, self.data, self.selected_column_name)
       ```

    4. Update your processing functions (e.g., `generate_answers`, `create_embeddings`,
       `execute_scripts_in_table`) to accept the column name as an argument.

    5. Remove any `uic.loadUi(self.gui, self)` calls from your widget, as the UI is already loaded automatically by `BaseListWidget`.

    6. If you only need "Data" as input and "Data" as output, remove the Inputs, Outputs and set() methods from your widget.
       They are already defined in BaseListWidget.

       ⚠️ Optional: Only if you truly need to customize input handling, you can
       override the `set_data()` method. In that case, make sure to update
       the `var_selector` (a `FilterableVariableList`) like this:

       ```python
       self.var_selector.add_variables(self.data.domain)
       self.var_selector.select_variable_by_name(self.selected_column_name)
       ```
    7. You can use
        autorun = Setting(True)
        and self.autorun = True / self.autorun = False
        to control whether the process is triggered automatically when input data is received.

    8.self.data is the input data stream. Make sure to create
        new_data = self.data.copy()
        to avoid issues when the process is triggered several time via a push button.
    """
    # Settings
    selected_column_name = Setting("Default")  # set the targeted column by default as "Default"

    class Inputs:
        data = Input("Data", Orange.data.Table)

    class Outputs:
        data = Output("Data", Orange.data.Table)

    @Inputs.data
    def set_data(self, in_data):
        self.data = in_data
        if self.data:
            self.var_selector.add_variables(self.data.domain)
            self.var_selector.select_variable_by_name(self.selected_column_name)
        if self.autorun:
            self.run()


    def __init__(self):
        super().__init__()
        # Load the .ui file if defined
        if self.gui and os.path.exists(self.gui):
            uic.loadUi(self.gui, self)

        # --- Insert filterable variable list ---
        self.var_selector = FilterableVariableList(self)
        placeholder = self.findChild(QWidget, "placeholder")  # add a QWidget in your .ui
        if placeholder:
            self.var_selector.setParent(placeholder)
            self.var_selector.setGeometry(0, 0, placeholder.width(), placeholder.height())
            self.var_selector.show()
        self.var_selector.selected_variable.connect(self.on_variable_selected)

        # Data management
        self.data = None
        # If self.autorun exists, it is preserved. Otherwise, it is set to True.
        self.autorun = getattr(self, "autorun", True)

    def on_variable_selected(self, var_name):
        """Update the selected column when the user clicks an item."""
        self.selected_column_name = var_name




# Widget Qt pour sélectionner des variables dans un widget Orange
class FilterableVariableList(QWidget):
    """
    A scrollable list of Orange variables with:
      - Single selection
      - Colored icons based on type
      - Filter field at the top
    """
    selected_variable = pyqtSignal(str)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.all_vars = []

        # Layout
        self.layout = QVBoxLayout(self)
        self.filter_edit = QLineEdit()
        self.filter_edit.setPlaceholderText("Filter variables...")
        self.layout.addWidget(self.filter_edit)

        self.list_widget = QListWidget()
        self.list_widget.setSelectionMode(QListWidget.SingleSelection)
        self.layout.addWidget(self.list_widget)

        # Connect filter
        self.filter_edit.textChanged.connect(self.apply_filter)
        # Connect variable selection
        self.list_widget.itemSelectionChanged.connect(self.on_selection_changed)


    def add_variables(self, variables):
        """Store all variables and populate the list"""
        self.all_vars = variables
        self.populate_list(self.all_vars)

    def populate_list(self, variables):
        """Populate the QListWidget with items and icons using Orange standard icons"""
        self.list_widget.clear()
        if variables is not None:
            for var in variables:
                item = QListWidgetItem(var.name)
                item.setIcon(gui.attributeIconDict[var])  # <-- standard Orange icon
                self.list_widget.addItem(item)

    def apply_filter(self, text):
        """Filter the list based on input"""
        text = text.lower()
        filtered = [v for v in self.all_vars if text in v.name.lower()]
        self.populate_list(filtered)

    def on_selection_changed(self):
        selected = self.get_selected_name()
        if selected:
            self.selected_variable.emit(selected)

    def get_selected_name(self):
        """Get the selected variable's name"""
        selected_items = self.list_widget.selectedItems()
        return selected_items[0].text() if selected_items else None

    def select_variable_by_name(self, name):
        """Select an item in the list by its variable name."""
        for i in range(self.list_widget.count()):
            item = self.list_widget.item(i)
            if item.text() == name:
                self.list_widget.setCurrentItem(item)
                self.list_widget.scrollToItem(item)  # optional: ensure it's visible
                return



import os
import sys
import re
import copy
import numpy as np
import ntpath

import Orange.data
from Orange.data import StringVariable, Domain, Table
from AnyQt.QtWidgets import QApplication
from Orange.widgets import widget
from Orange.widgets.utils.signals import Input, Output
from sentence_transformers import SentenceTransformer


if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.llm import lmstudio
    from Orange.widgets.orangecontrib.AAIT.utils import thread_management
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.llm import lmstudio
    from orangecontrib.AAIT.utils import thread_management
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file


@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWGenerateQuestions(widget.OWWidget):
    name = "Generate Questions"
    description = "Generate questions on the column 'content' of an input table"
    category = "AAIT - LLM INTEGRATION"
    icon = "icons/owgeneratequestions.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owgeneratequestions.svg"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owgeneratequestions.ui")
    want_control_area = False
    priority = 1060

    class Inputs:
        #data = Input("Data for GPT4All", Orange.data.Table)
        data_lmstudio = Input("Data for LMStudio", Orange.data.Table)
        model = Input("Model", str, auto_summary=False)
        embedder = Input("Embedder", SentenceTransformer, auto_summary=False)

    class Outputs:
        data = Output("Data", Orange.data.Table)


    @Inputs.data_lmstudio
    def set_data_lmstudio(self, in_data):
        if in_data is None:
            self.Outputs.data.send(None)
            return
        self.data_lmstudio = in_data
        self.data = None
        if self.autorun:
            self.run()

    @Inputs.model
    def set_model(self, in_model):
        self.model = in_model
        if self.autorun:
            self.run()

    @Inputs.embedder
    def set_embedder(self, in_embedder):
        self.embedder = in_embedder
        if self.autorun:
            self.run()

    def __init__(self):
        super().__init__()
        # Qt Management
        self.setFixedWidth(470)
        self.setFixedHeight(300)
        uic.loadUi(self.gui, self)

        # Data Management
        self.data = None
        self.data_lmstudio = None
        self.model = None
        self.embedder = None
        self.thread = None
        self.autorun = True
        self.result = None
        self.post_initialized()

    def run(self):
        self.error("")
        # if thread is running quit
        if self.thread is not None:
            self.thread.safe_quit()
            return

        # Verify inputs
        if self.data is None and self.data_lmstudio is None:
            return

        if self.data and self.data_lmstudio:
            self.error('You cannot have both inputs GPT4All and LMStudio.')
            return

        if self.model is None:
            return

        if self.embedder is None:
            return

        # If LMStudio is used : verify settings
        if self.data_lmstudio is not None:
            llm = 'lmstudio'
            data = self.data_lmstudio
            model_list = lmstudio.get_model_lmstudio()
            if isinstance(model_list, str) and model_list == "Error":
                self.error("Please launch the LMStudio app and start the server.")
                return
            # Verify that the models directory is set in LMStudio (aait_store)
            if model_list is None:
                self.error("Your models directory in LMStudio is not set properly.")
                return

            # Verify that the plugged model exists in your LMStudio model list
            if not any(ntpath.basename(d["id"]).lower() == ntpath.basename(self.model.lower()) for d in model_list["data"]):
                self.error(f"Model not found in LMStudio. You are trying to use {ntpath.basename(self.model.lower())}, please verify that this model's API identifier exists in LMStudio.")
                return
            for d in model_list["data"]:
                if ntpath.basename(d["id"]).lower() == ntpath.basename(self.model.lower()):
                    self.model = d["id"]

        # IF GPT4All is used
        if self.data is not None:
            llm = 'gpt4all'
            data = self.data
            self.model = ntpath.basename(self.model)

        # Verification of in_data
        try:
            data.domain["content"]
        except KeyError:
            self.error('You need a "content" column in input data')
            return

        if type(data.domain["content"]).__name__ != 'StringVariable':
            self.error('"content" column needs to be a Text')
            return

        # Start progress bar
        self.progressBarInit()

        # Connect and start thread : main function, progress, result and finish
        # --> progress is used in the main function to track progress (with a callback)
        # --> result is used to collect the result from main function
        # --> finish is just an empty signal to indicate that the thread is finished
        self.thread = thread_management.Thread(self.generate_questions_on_table, data, self.model, self.embedder, llm)
        self.thread.progress.connect(self.handle_progress)
        self.thread.result.connect(self.handle_result)
        self.thread.finish.connect(self.handle_finish)
        self.thread.start()

    def handle_progress(self, value: float) -> None:
        self.progressBarSet(value)

    def handle_result(self, result):
        try:
            self.result = result
            self.Outputs.data.send(result)
        except Exception as e:
            print("An error occurred when sending out_data:", e)
            self.Outputs.data.send(None)
            return

    def handle_finish(self):
        print("Question generation finished")
        self.progressBarFinished()
        self.thread = None

    def post_initialized(self):
        pass


    def generate_questions_on_table(self, table, model, embedder, llm="gpt4all", progress_callback=None, argself=None):
        data = copy.deepcopy(table)
        new_metas = list(data.domain.metas) + [StringVariable("Questions")]
        new_domain = Domain(data.domain.attributes, data.domain.class_vars, new_metas)

        new_rows = []
        for i, row in enumerate(data):
            print(f"\n\n -- Working on row {i}")
            content = row["content"].value
            questions = self.generate_dataset_on_chunk(content, model, embedder, llm)
            for question in questions:
                new_metas_values = list(row.metas) + [question]
                new_instance = Orange.data.Instance(new_domain, [row[x] for x in data.domain.attributes] + [row[y] for y in data.domain.class_vars] + new_metas_values)
                new_rows.append(new_instance)
                if progress_callback is not None:
                    progress_value = float(100 * (i + 1) / len(data))
                    progress_callback(progress_value)
            if argself:
                if argself.stop:
                    break

        out_data = Table.from_list(domain=new_domain, rows=new_rows)
        return out_data



    def generate_dataset_on_chunk(self, chunk, model, embedder, llm="gpt4all", threshold=0.7, max_repeat=5, safety_limit=15,
                                  previous_results=None):
        """
        Generates a dataset of unique questions and answers for a given text chunk.

        Parameters:
            model: The language model used to generate questions and answers.
            embedder: A sentence embedding model used to encode questions for similarity checks.
            chunk (str): The text chunk to process.
            threshold (float): The cosine similarity threshold for considering two questions as duplicates (default: 0.95).
            max_repeat (int): The maximum number of iterations to attempt generating unique questions (default: 20).
            safety_limit (int): A hard limit on the total number of iterations to avoid infinite loops (default: 200).
            previous_results(dict): A dictionnary containing the lists of embeddings, questions and answers for the given chunk

        Returns:
            tuple: A tuple containing two lists:
                - saved_questions (list of str): The list of unique questions.
                - saved_answers (list of str): The list of answers corresponding to the questions.
        """
        # Initialize lists to store unique embeddings, questions, and answers
        saved = []  # Stores embeddings of unique questions
        saved_questions = []  # Stores unique questions
        saved_answers = []  # Stores answers corresponding to the unique questions

        # Initialize counters for loop control
        repeat = 0  # Tracks the number of successful iterations adding unique questions
        safety = 0  # Tracks the total number of iterations to prevent infinite loops

        # Continue generating questions until max_repeat or safety_limit is reached
        while repeat < max_repeat and safety < safety_limit:
            print(f"\nRepeated: {repeat}/{max_repeat} | Safety: {safety}/{safety_limit}")
            # Generate questions and answers using the model
            try:
                model_output = self.generate_questions_and_answers(model, chunk, llm)
                questions = self.parse_questions_and_answers(model_output)
            except TypeError:
                self.error("The model could not generate an answer. Please make sure your model is loaded in LMStudio.")
                return []
            print("### Questions generated:", len(questions))
            answers = questions

            if questions != [] and answers != []:
                # Iterate over the generated questions and answers
                added = []
                # print(f"{len(questions)} questions generated")
                # print(f"{len(saved)} questions to compare")
                for i in range(len(questions)):
                    question = questions[i]
                    answer = answers[i]
                    to_add = True  # Flag to indicate if the current question is unique

                    # Encode the current question into embeddings
                    embeddings = embedder.encode(question, show_progress_bar=False)

                    # Check the similarity of the current question with previously saved questions
                    for saved_embeddings in saved:
                        cos_sim = cosine_similarity(embeddings, saved_embeddings)
                        if cos_sim > threshold:  # If too similar, mark for exclusion
                            to_add = False
                            # print("---> Duplicate !", cos_sim)
                            break  # No need to check further if similarity is already too high

                    # Add the question and its answer if it is unique
                    if to_add:
                        saved.append(embeddings)  # Save the embeddings of the unique question
                        saved_questions.append(question)  # Save the question
                        saved_answers.append(answer)  # Save the corresponding answer

                    added.append(to_add)
                print("### Questions stored:", sum(added))
                # print(f"Added {sum(added)} questions to the question list ! --> {len(saved)} questions saved\n")
                if not any(added):
                    repeat += 1

            # Increment the safety counter for every loop iteration
            safety += 1

        # Return the collected unique questions and their corresponding answers
        return saved_questions


    def generate_questions_and_answers(self, model, text, llm="gpt4all", max_tokens=4096, temperature=1, top_p=1, top_k=50, stop_tokens=None):
        """
        Generates question-answer pairs based on a provided text using a language model (LLM).

        Parameters:
            model (Llama): The language model instance with a callable API for generating text.
            text (str): The input document or excerpt to generate questions and answers from.
            llm (str): The type of backend to use, "gpt4all" or "lmstudio".
            max_tokens (int): Maximum tokens for the model's output (default is 4096).
            temperature (float): Sampling temperature to control randomness (default is 1).
            top_p (float): Top-p sampling to limit the token pool by probability (default is 1).
            top_k (int): Top-k sampling to limit the token pool by rank (default is 50).
            stop_tokens (list): List of stop sequences to terminate generation (default is None).

        Returns:
            str: The generated question-answer pairs.
        """
        # Define the prompt
        prompt = f"""### Contexte : Tu es un examinateur. Ton objectif est de préparer des questions pour un examen. Les candidats vont être interrogés sur un document, dont ils ont dû apprendre le contenu. Ils n'ont aucun accès au document durant l'examen.


### Instructions :
- Génère des questions simples et précises qui reflètent les informations principales du document. 
- Les candidats n'auront pas accès au document pour répondre aux questions.
- Tu ne dois JAMAIS faire d'allusion au document
- Tu ne dois JAMAIS utiliser la formule "D'après ce document".
- Tu ne dois JAMAIS utiliser la formule "Comme mentionne ce document".
- Tu ne dois JAMAIS utiliser la formule "Selon ce document".
- Respecte la stucture suivante :
1)
2)
3)
etc...


### Document :



"{text}"




### Assistant :"""
        # if llm == "gpt4all":
        #     try:
        #         # Generate output using the LLM
        #         localhost = "localhost:4891"
        #         response = GPT4ALL.call_completion_api(localhost=localhost, message_content=prompt, model_name=model,
        #                                                temperature=1, top_p=1)
        #         answer = GPT4ALL.clean_response(response)
        #
        #
        #         # Extract and return the text portion of the response
        #         return answer.strip()
        #     except Exception as e:
        #         return f"Error generating questions and answers: {e}"

        if llm == "lmstudio":
            try:
                # Generate output using the LLM
                response = lmstudio.appel_lmstudio(prompt, model)
                return response
            except Exception as e:
                return f"Error generating questions and answers: {e}"


    def parse_questions_and_answers(self, qa_text):
        """
        Parses a string containing questions and answers into a structured list of dictionaries.

        Parameters:
            qa_text (str): The input string containing questions and answers in the format:
                           Q: "Question..."
                           R: "Answer..."

        Returns:
            list: A list of dictionaries, each containing a 'question' and 'answer' key.
        """
        # Regular expressions to match questions and answers
        question_pattern = r"\d+[.)]\s?(.*?)(?:\n|$)"

        # Extract questions and answers using regex
        questions = re.findall(question_pattern, qa_text)

        # Ensure equal number of questions and answers
        return questions

def cosine_similarity(u, v):
    dot_product = np.dot(u, v)
    norm_u = np.linalg.norm(u)
    norm_v = np.linalg.norm(v)
    return dot_product / (norm_u * norm_v)


if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWGenerateQuestions()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()



import os
import sys

from AnyQt.QtWidgets import QApplication
from Orange.widgets import widget
from Orange.widgets.utils.signals import Output
from AnyQt.QtWidgets import QLineEdit
from sentence_transformers import SentenceTransformer
from Orange.widgets.settings import Setting


if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils import MetManagement
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils import MetManagement
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file


@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWModelEmbeddings(widget.OWWidget):
    name = "Model - Embeddings"
    description = "Load the embeddings model from the user's input path."
    category = "AAIT - MODELS"
    icon = "icons/owmodel_embeddings.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owmodel_embeddings.svg"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owmodel_embeddings.ui")
    priority = 1061
    want_control_area = False

    model_path = Setting("")

    class Outputs:
        out_model = Output("Model", SentenceTransformer, auto_summary=False)

    def __init__(self):
        super().__init__()
        # Qt Management
        self.setFixedWidth(470)
        self.setFixedHeight(300)
        uic.loadUi(self.gui, self)
        # TextEdit
        self.lineedit_path = self.findChild(QLineEdit, 'lineEdit')
        self.lineedit_path.setText(self.model_path)
        self.lineedit_path.editingFinished.connect(self.edit_model_path)

        # Load model
        self.run()


    def edit_model_path(self):
        path = self.lineedit_path.text().strip('"').strip("'")
        self.model_path = path
        self.run()


    def run(self):
        # Reset indications on widget
        self.error("")
        self.warning("")
        self.information("")

        # Load the model and send it
        model = self.load_model()
        self.Outputs.out_model.send(model)


    def load_model(self):
        path = self.model_path
        if not os.path.isabs(path):
            local_store_path = MetManagement.get_local_store_path()
            path = os.path.join(local_store_path, "Models", "NLP", path)

        if not os.path.isdir(path):
            self.error(f"Model directory does not exist: {path}")
            return None

        try:
            model = SentenceTransformer(path, device="cpu")
            return model
        except Exception as e:
            self.error(f"The model could not be loaded: {e}")
            return


if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWModelEmbeddings()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()



import os
import Orange
from Orange.data import Domain, Table, StringVariable, ContinuousVariable
from sentence_transformers import SentenceTransformer

### Chonkie
from chonkie import TokenChunker, SentenceChunker, RecursiveChunker, SemanticChunker, LateChunker, CodeChunker
if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.llm import wordchunker_deprecated
    from Orange.widgets.orangecontrib.AAIT.utils.MetManagement import get_local_store_path
else:
    from orangecontrib.AAIT.llm import wordchunker_deprecated
    from orangecontrib.AAIT.utils.MetManagement import get_local_store_path


def create_chunks(table, column_name, tokenizer="character", chunk_size=300, chunk_overlap=100, mode="tokens",
                  progress_callback=None, argself=None):
    """
    Chunk the text in `column_name` of an Orange Table using a specialized chunker.

    Splits each row's text into chunks based on the selected mode (Token, Sentence,
    Recursive, or Markdown). Adds the chunked text and its metadata as new meta
    columns to the table.

    Parameters:
        table (Table): Input data table.
        column_name (str): Name of the text column to chunk.
        tokenizer (str): Tokenizer type (e.g., "character").
        chunk_size (int): Target chunk size.
        chunk_overlap (int): Overlap between chunks (not used in all modes).
        mode (str): Chunking strategy ("Token", "Sentence", "Recursive", "Markdown").
        progress_callback (callable): Optional progress reporter.
        argself: Optional caller reference.

    Returns:
        Table: The table with added meta columns: "Chunks", "Chunks size", and "Metadata".
    """
    print("This widget is being updated : default tokenizer 'character' enabled for compatibility !!")
    tokenizer = "character"
    model_name = "character"

    # Définir la fonction de chunking selon le mode
    if mode == "tokens":
        chunker = TokenChunker(tokenizer=tokenizer, chunk_size=chunk_size, chunk_overlap=chunk_overlap)
    elif mode == "words":
        path_ugly = os.path.join(get_local_store_path(), "Models", "NLP", "all-mpnet-base-v2")
        tokenizer = SentenceTransformer(path_ugly, device="cpu")
        model_name = "MPNET"
        chunker = wordchunker_deprecated.WordChunker(tokenizer=tokenizer, chunk_size=chunk_size, chunk_overlap=chunk_overlap)
    elif mode == "sentence":
        chunker = SentenceChunker(tokenizer=tokenizer, chunk_size=chunk_size, chunk_overlap=chunk_overlap,
                                  min_sentences_per_chunk=1)
    elif mode == "markdown":
        markdown_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources/markdown_recipe.json")
        chunker = RecursiveChunker.from_recipe(path=markdown_path,
                                               tokenizer=tokenizer,
                                               chunk_size=chunk_size,
                                               min_characters_per_chunk=1)

    # TODO : Ajouter la gestion des paramètres dans le .ui
    # Nécessite des "rules" pour faire un chunking différent de Token ou Sentence
    elif mode == "Recursive":
        chunker = RecursiveChunker(tokenizer=tokenizer, chunk_size=chunk_size, min_characters_per_chunk=24)
    # À tester avant d'ajouter la fonctionnalité dans l'UI
    # Model d'embeddings REQUIS !
    elif mode == "Semantic":
        chunker = SemanticChunker(embedding_model=tokenizer, threshold=0.7, chunk_size=chunk_size, similarity_window=3)
    # À tester avant d'ajouter la fonctionnalité dans l'UI
    # Model d'embeddings REQUIS !
    elif mode == "Late":
        chunker = LateChunker(embedding_model=tokenizer, chunk_size=chunk_size, min_characters_per_chunk=24)
    elif mode == "Code":
        chunker = CodeChunker("blabla")
    else:
        raise ValueError(f"Invalid mode: {mode}. Valid modes are: Token, Sentence, Recursive, Markdown")

    new_metas = list(table.domain.metas) + [StringVariable("Chunks"),
                                            ContinuousVariable("Chunks size"),
                                            ContinuousVariable("Chunks index"),
                                            StringVariable("Metadata")]
    metas_infos = [ContinuousVariable("Overlap"),
                   StringVariable("Chunking model"),
                   StringVariable("Chunking method")]

    new_domain = Domain(table.domain.attributes, table.domain.class_vars, new_metas)
    new_domain_info = Domain([], metas=metas_infos)

    new_rows = []
    info_rows = []
    for i, row in enumerate(table):
        content = row[column_name].value
        chunks = chunker(content)
        # For each chunk in the chunked data
        for j, chunk in enumerate(chunks):
            # Build new metas with previous data and the chunk
            new_metas_values = list(row.metas) + [chunk.text,
                                                  chunk.token_count,
                                                  j, # Chunks index
                                                  ""]
            info_rows.append([chunk_overlap, model_name, mode])
            # Create the new row instance
            new_instance = Orange.data.Instance(new_domain,
                                                [row[x] for x in table.domain.attributes] + [row[y] for y in
                                                                                             table.domain.class_vars] + new_metas_values)
            # Store the new row
            new_rows.append(new_instance)

        if progress_callback is not None:
            progress_value = float(100 * (i + 1) / len(table))
            progress_callback(progress_value)
        if argself is not None:
            if argself.stop:
                break

    return Table.from_list(domain=new_domain, rows=new_rows), Table.from_list(domain=new_domain_info, rows=info_rows)


# https://orange3.readthedocs.io/projects/orange-development/en/latest/tutorial-settings.html


import ctypes
import os
import sys

import Orange.data
from AnyQt.QtCore import QTimer
from AnyQt.QtWidgets import QButtonGroup, QLabel, QRadioButton
from Orange.widgets import  widget
from Orange.widgets.settings import Setting
from Orange.widgets.widget import Input, Output


class OWOptimisationSelection(widget.OWWidget):
    name = "Optimization - Selection"
    description = "Choose whether you want Optuna to run automatically or not"
    icon = "icons/owoptimisationselection.png"
    category = "AAIT - ALGORITHM"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owoptimisationselection.png"

    priority = 1022
    want_control_area = False
    dossier_du_script = os.path.dirname(os.path.abspath(__file__))
    # Les paramètres du widget
    radio_choice = Setting(2)  # Pour stocker l'index de la sélection
    in_pointer = None

    class Inputs:
        data = Input("Data", Orange.data.Table)
        in_pointer = Input("Get study for the loop", str, auto_summary=False)  # sended as the widget is created

    class Outputs:
        data = Output("Data", Orange.data.Table)

    @Inputs.in_pointer
    def set_pointer(self, dataset):
        if dataset:
            self.in_pointer = int(dataset)

    @Inputs.data
    def set_data(self, in_data):
        self.data = []
        data = in_data
        column_name = "Source ID"
        # Vérifier si la colonne existe dans le domaine des attributs
        column_exists = column_name in [attr.name for attr in data.domain]
        if column_exists:
            column_index = data.domain.index(column_name)
            d = []
            for i in range(len(data)):
                if data[i][column_index] == "created":
                    d = data[i]
            self.data = Orange.data.Table(data.domain, [d])
        else:
            self.data = in_data

        if self.in_pointer is not None:
            if len(self.data) == 1:
                ctypes.cast(self.in_pointer, ctypes.py_object).value.change_last_ligne(self.data)

        if self.radio_choice == 0:
            self.Outputs.data.send(self.data)


    def __init__(self):
        super().__init__()

        self.data = None

        # Utiliser self.controlArea.layout() pour s'assurer que les widgets sont ajoutés à la zone de contrôle
        self.label = QLabel("Choose an option :")
        self.mainArea.layout().addWidget(self.label)

        # Créer un groupe de boutons radio
        self.radio_group = QButtonGroup(self)

        # Bouton radio 1
        self.radio1 = QRadioButton("Automatically")
        self.radio1.setChecked(self.radio_choice == 0)  # Cocher par défaut si radio_choice est 0
        self.radio1.toggled.connect(self.on_radio_changed)
        self.mainArea.layout().addWidget(self.radio1)
        self.radio_group.addButton(self.radio1, 0)

        # Bouton radio 2
        self.radio2 = QRadioButton("Manually")
        self.radio2.setChecked(self.radio_choice == 1)  # Cocher si radio_choice est 1
        self.radio2.toggled.connect(self.on_radio_changed)
        self.mainArea.layout().addWidget(self.radio2)
        self.radio_group.addButton(self.radio2, 1)

        # Bouton radio 3
        self.radio3 = QRadioButton("No data")
        self.radio3.setChecked(self.radio_choice == 2)  # Cocher si radio_choice est 1
        self.radio3.toggled.connect(self.on_radio_changed)
        self.mainArea.layout().addWidget(self.radio3)
        self.radio_group.addButton(self.radio3, 2)

        self.setFixedWidth(480)
        self.setFixedHeight(354)
        self.setAutoFillBackground(True)

        # Initialiser le timer pour un délai de 5 secondes
        self.timer = QTimer(self)
        self.timer.setSingleShot(True)  # S'assurer que le timer ne se déclenche qu'une fois
        self.timer.timeout.connect(self.select_radio3)



    def on_radio_changed(self):
        # Détecte quel bouton est sélectionné
        selected_button = self.radio_group.checkedId()
        if selected_button == 0:
            self.Outputs.data.send(self.data)
        elif selected_button == 1:
            self.Outputs.data.send(self.data)
            self.timer.start(200)
        elif selected_button == 2:
            print("No data")
        # Mise à jour du paramètre de sauvegarde
        self.radio_choice = selected_button


    def select_radio3(self):
        # Sélectionner le bouton radio 2 après 5 secondes
        self.radio3.setChecked(True)

if __name__ == "__main__":
    from AnyQt.QtWidgets import QApplication

    app = QApplication(sys.argv)
    mon_objet = OWOptimisationSelection()
    mon_objet.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()





from AnyQt.QtWidgets import QApplication
from AnyQt.QtCore import QTimer
import re
#allows displaying a widget on opening
# add (not sure!)
# save_position = False
# after class declaration
# call show_and_adjust_at_opening at the end of init
def show_and_adjust_at_opening(argself,str_position):
    # "left", "top-left", "bottom-left","bottom", "bottom-left", "bottom-right"
    str_position=str(str_position)
    if str_position=="None" or str_position=="none":
        return
    argself.show()
    QTimer.singleShot(1000, lambda: adjust_to_quarter(argself,str_position))

def _parse_variable_position(spec: str):
    """
    Attend: vXX_dvYY_hxx_dhyy
      vXX  = % de l'écran (Y du coin haut)
      dvYY = % de l'écran (hauteur)
      hxx  = % de l'écran (X du coin gauche)
      dhyy = % de l'écran (largeur)
    Retourne (v, dv, h, dh) en int si OK, sinon None.
    """
    if not isinstance(spec, str):
        return None

    m = re.fullmatch(r"v(\d{1,3})_dv(\d{1,3})_h(\d{1,3})_dh(\d{1,3})", spec.strip())
    if not m:
        return None

    v, dv, h, dh = map(int, m.groups())

    # bornes simples: 0..100
    if any(p < 0 or p > 100 for p in (v, dv, h, dh)):
        return None

    # éviter taille nulle (tu peux enlever si tu veux autoriser 0)
    if dv == 0 or dh == 0:
        return None

    # vérifier que le rectangle reste dans l'écran (en %)
    if v + dv > 100 or h + dh > 100:
        return None

    return v, dv, h, dh

def adjust_to_quarter(argself, position: str):
    """
    Resize and position the widget depending on a screen region:
    - left / right  → full height, half width
    - top / bottom  → full width, half height
    - corners       → quarter (half width × half height)
    """
    app = QApplication.instance()
    screen = app.primaryScreen().availableGeometry()

    sw, sh = screen.width(), screen.height()

    parsed = _parse_variable_position(position)
    if parsed is not None:
        v, dv, h, dh = parsed

        # conversion % -> pixels
        y = screen.y() + int(round(sh * (v / 100.0)))
        h_px = int(round(sh * (dv / 100.0)))

        x = screen.x() + int(round(sw * (h / 100.0)))
        w_px = int(round(sw * (dh / 100.0)))

        argself.resize(w_px, h_px)
        argself.move(x, y)
        return



    if position == "fullscreen":
        # full screen
        w, h = sw, sh
    # ---- CALCUL TAILLE ----
    elif position in ("left", "right"):
        # full height, half width
        w, h = sw // 2, sh

    elif position in ("top", "bottom"):
        # full width, half height
        w, h = sw, sh // 2

    elif position in ("top-left", "top-right", "bottom-left", "bottom-right"):
        # quarter (default case)
        w, h = sw // 2, sh // 2
    elif position == "none" or position == "None":
        return  # ou ne rien faire
    else:
        # fallback: centered
        w, h = sw // 2, sh // 2

    # ---- CALCUL POSITION ----
    if position in ("left", "top-left", "bottom-left"):
        x = screen.x()
    elif position in ("right", "top-right", "bottom-right"):
        x = screen.x() + sw - w
    else:
        x = screen.x() + (sw - w) // 2

    if position in ("top", "top-left", "top-right"):
        y = screen.y()
    elif position in ("bottom", "bottom-left", "bottom-right"):
        y = screen.y() + sh - h
    else:
        y = screen.y() + (sh - h) // 2

    # ---- APPLIQUER ----
    argself.resize(w, h)
    argself.move(x, y)




# OWSplitPath.py
from __future__ import annotations

from pathlib import Path
from typing import Tuple
import sys
import numpy as np
from AnyQt.QtWidgets import QApplication, QCheckBox
from Orange.data import StringVariable
from Orange.widgets.settings import Setting
import os
if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
    from Orange.widgets.orangecontrib.AAIT.utils import base_widget
else:
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
    from orangecontrib.AAIT.utils import  base_widget





# =========================
# Widget
# =========================
@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWSplitPath(base_widget.BaseListWidget):
    name = "Split Path"
    description = "Normalize path (resolve) and split into directory and file name."
    icon = "icons/split_path.png"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/split_path.png"
    category = "AAIT - TOOLBOX"

    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owsplitpath.ui")
    want_control_area = False
    priority = 100
    # Settings
    autorun = Setting(True)
    with_file_name_extension = Setting(False)
    with_file_name_without_extension = Setting(False)
    with_parent_dir_name_column = Setting(False)


    def _safe_str(self,x) -> str:
        if x is None:
            return ""
        try:
            if isinstance(x, float) and np.isnan(x):
                return ""
        except Exception:
            pass
        return str(x)

    def normalize_path(self,p: str) -> str:
        """Normalize a path using pathlib.resolve(strict=False)."""
        s = self._safe_str(p).strip()
        if not s:
            return ""

        try:
            return Path(s).expanduser().resolve(strict=False).as_posix()
        except Exception:
            # fallback: return raw string if pathlib fails
            return s

    def split_dir_file(self,p: str) -> Tuple[str, str]:
        if not p:
            return "", ""
        try:
            pp = Path(p)
            return pp.parent.as_posix(), pp.name
        except Exception:
            return "", p

    def _split_name_parts(self, filename: str) -> Tuple[str, str]:
        """
        Retourne (stem, extension_sans_point).
        Ex: "tata.txt" -> ("tata", "txt")
            "archive.tar.gz" -> ("archive.tar", "gz")
            ".gitignore" -> (".gitignore", "")
        """
        if not filename:
            return "", ""
        try:
            p = Path(filename)
            ext = p.suffix[1:] if p.suffix.startswith(".") else p.suffix  # sans le point
            return p.stem, ext
        except Exception:
            # fallback simple
            if "." in filename and not filename.startswith("."):
                base, ext = filename.rsplit(".", 1)
                return base, ext
            return filename, ""

    def _parent_dir_name(self, directory: str) -> str:
        """
        Retourne le nom du dossier parent final.
        Ex: "C:/toto/titi" -> "titi"
        """
        if not directory:
            return ""
        try:
            return Path(directory).name
        except Exception:
            return directory.rstrip("/\\").split("/")[-1].split("\\")[-1]

    def __init__(self):
        super().__init__()
        self.setFixedWidth(470)
        self.setFixedHeight(500)
        self.data = None
        self.checkbox_mode = self.findChild(QCheckBox, 'checkBox_send')


        self.checkBox_file_name_extension= self.findChild(QCheckBox, 'checkBox_file_name_extension')
        self.checkBox_file_name_without_extension= self.findChild(QCheckBox, 'checkBox_file_name_without_extension')
        self.checkBox_parent_dir_name_column= self.findChild(QCheckBox, 'checkBox_parent_dir_name_column')

        self.checkbox_mode.setChecked(self.autorun)



        self.checkBox_file_name_extension.setChecked(self.with_file_name_extension)
        self.checkBox_file_name_without_extension.setChecked(self.with_file_name_without_extension)
        self.checkBox_parent_dir_name_column.setChecked(self.with_parent_dir_name_column)



        self.checkbox_mode.stateChanged.connect(self.toogle_persistent)
        self.checkBox_file_name_extension.stateChanged.connect(self.toogle_name_extension)
        self.checkBox_file_name_without_extension.stateChanged.connect(self.toogle_file_name_without_extension)
        self.checkBox_parent_dir_name_column.stateChanged.connect(self.toogle_parent_dir_name_column)




        self.pushButton_send.clicked.connect(self.run)

    def run(self):
        self.warning("")
        self.error("")

        if self.data is None:
            self.Outputs.data.send(None)
            return
        if not self.selected_column_name in self.data.domain:
            self.warning(f'Previously selected column "{self.selected_column_name}" does not exist in your data.')
            return

        # colonnes toujours ajoutées
        if "path_dir" in self.data.domain:
            self.error('cannot be used with "path_dir" in input')
            return
        if "file_name" in self.data.domain:
            self.error('cannot be used with "file_name" in input')
            return

        # colonnes optionnelles
        if self.with_file_name_extension and "file_name_extension" in self.data.domain:
            self.error('cannot be used with "file_name_extension" in input')
            return
        if self.with_file_name_without_extension and "file_name_without_extension" in self.data.domain:
            self.error('cannot be used with "file_name_without_extension" in input')
            return
        if self.with_parent_dir_name_column and "parent_dir_name" in self.data.domain:
            self.error('cannot be used with "parent_dir_name" in input')
            return

        list_dir = []
        name_path = []
        list_ext = []  # sans point
        list_stem = []  # sans extension
        list_parent = []  # nom du dossier parent

        for row in self.data:
            current_path = str(row[self.selected_column_name].value)
            raw = self._safe_str(current_path)
            norm = self.normalize_path(raw)
            d, name = self.split_dir_file(norm)

            list_dir.append(d)
            name_path.append(name)

            stem, ext = self._split_name_parts(name)
            if self.with_file_name_extension:
                list_ext.append(ext)
            if self.with_file_name_without_extension:
                list_stem.append(stem)
            if self.with_parent_dir_name_column:
                list_parent.append(self._parent_dir_name(d))

        out_put_data = self.data.copy()

        # toujours
        out_put_data = out_put_data.add_column(StringVariable("path_dir"), list_dir, to_metas=True)
        out_put_data = out_put_data.add_column(StringVariable("file_name"), name_path, to_metas=True)

        # optionnel
        if self.with_file_name_extension:
            out_put_data = out_put_data.add_column(StringVariable("file_name_extension"), list_ext, to_metas=True)

        if self.with_file_name_without_extension:
            out_put_data = out_put_data.add_column(StringVariable("file_name_without_extension"), list_stem,
                                                   to_metas=True)

        if self.with_parent_dir_name_column:
            out_put_data = out_put_data.add_column(StringVariable("parent_dir_name"), list_parent, to_metas=True)

        self.Outputs.data.send(out_put_data)

    def toogle_persistent(self, enabled):
        if self.checkbox_mode.isChecked():
            self.autorun = True
        else:
            self.autorun = False

    def toogle_name_extension(self, enabled):
        if self.checkBox_file_name_extension.isChecked():
            self.with_file_name_extension = True
        else:
            self.with_file_name_extension = False

    def toogle_file_name_without_extension(self, enabled):
        if self.checkBox_file_name_without_extension.isChecked():
            self.with_file_name_without_extension = True
        else:
            self.with_file_name_without_extension = False


    def toogle_parent_dir_name_column(self, enabled):
        if self.checkBox_parent_dir_name_column.isChecked():
            self.with_parent_dir_name_column = True
        else:
            self.with_parent_dir_name_column = False


if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWSplitPath()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()



import os
import sys

from AnyQt.QtWidgets import QApplication
from Orange.data import Table, Domain, StringVariable
from Orange.widgets.utils.signals import Output, Input
from Orange.widgets import widget

if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.llm import process_documents
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
else:
    from orangecontrib.AAIT.llm import process_documents
    from orangecontrib.AAIT.utils.import_uic import uic



class OWProcessDocumentsFromPath(widget.OWWidget): # type: ignore
    name = "Process Documents From Path"
    description = ("Read the documents from a folder and search for a 'embeddings.pkl' file. Then the content of "
                   "'embeddings.pkl' is returned as Processed Data, while the content of the documents (not contained "
                   "in 'embeddings.pkl') is returned as Data")
    icon = "icons/processdocuments.svg"
    category = "AAIT - TOOLBOX"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/processdocuments.svg"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owprocessdocuments.ui")
    priority = 1211
    category = "Advanced Artificial Intelligence Tools"
    want_control_area = False

    class Inputs:
        dirpath = Input("Path", str, auto_summary=False)
        path_table = Input("Path Table", Table)

    class Outputs:
        data = Output("Data", Table)
        processed_data = Output("Processed Data", Table)


    @Inputs.dirpath
    def set_dirpath(self, in_dirpath):
        if in_dirpath is not None:
            self.dirpath = in_dirpath
            self.run()

    @Inputs.path_table
    def set_path_table(self, in_path_table):
        if in_path_table is not None:
            if "path" in in_path_table.domain:
                self.dirpath = in_path_table[0]["path"].value
                self.run()
            else:
                self.warning("You need a 'path' variable from which the data will be loaded.")


    def __init__(self):
        super().__init__()
        # Qt Management
        self.setFixedWidth(470)
        self.setFixedHeight(300)
        uic.loadUi(self.gui, self)

        # Data Management
        self.dirpath = None

    def run(self):
        self.error("")

        if self.dirpath is None:
            return

        if os.path.exists(self.dirpath):
            out_data, out_processed_data = process_documents.process_documents(self.dirpath)
            self.Outputs.data.send(out_data)
            self.Outputs.processed_data.send(out_processed_data)
        else:
            var_path = StringVariable("path")
            var_name = StringVariable("name")
            var_content = StringVariable("content")
            dom = Domain([], metas=[var_path, var_name, var_content])
            out_data = Table.from_list(dom, rows=[["", "", ""]])
            self.Outputs.data.send(out_data)
            self.Outputs.processed_data.send(out_data)
            self.error(f"Selected path '{self.dirpath}' does not exist.")

if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWProcessDocumentsFromPath()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()


import Orange
from packaging import version

import sys

import os
# def check_executable_path():
#     exe = sys.executable
#
#     # caractères spéciaux interdits (tu peux en ajouter)
#     forbidden = r'[^A-Za-z0-9_\-./:\\]'  # tout ce qui n'est PAS ce set
#     pattern = re.compile(forbidden)
#
#     if " " in exe or pattern.search(exe):
#         return False
#     return True
#
# if not check_executable_path():
#     if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
#         from orangecontrib.AAIT.utils import SimpleDialogQt
#     else:
#         from orangecontrib.AAIT.utils import SimpleDialogQt
#     SimpleDialogQt.BoxError("You must install this program in a path that does not contain spaces or special characters.")
#     exit(0)

def check_executable_length(max_length=260):
    exe = sys.executable
    length = len(exe)

    if length > max_length:
        print(f"ERROR: The installation path is too long ({length} characters). "
              f"Maximum allowed is {max_length}.")
        return False
    return True

if not check_executable_length():
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        from orangecontrib.AAIT.utils import SimpleDialogQt
    else:
        from orangecontrib.AAIT.utils import SimpleDialogQt
    SimpleDialogQt.BoxError("The installation path is too long. Please choose a shorter path..")
    exit(0)


target_version = version.parse("3.40")
current_version = version.parse(Orange.version.version)
if current_version < target_version: # Skip the file
    print("Orange version not compatible with all of AAIT functions !")
    SimpleDialogQt.BoxError("Orange version not compatible with all of AAIT functions !")

else: # Execute the file
    import os
    import tempfile
    import gc
    from orangewidget.workflow.discovery import WidgetDiscovery
    #from orangecanvas.registry import CategoryDescription
    from orangecanvas.registry.utils import category_from_package_globals
    from orangecanvas.utils.pkgmeta import get_distribution




    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\","/"):
         from Orange.widgets.orangecontrib.AAIT.utils.tools import change_owcorpus
         from Orange.widgets.orangecontrib.AAIT.utils import MetManagement
    else:
        from orangecontrib.AAIT.utils.tools import change_owcorpus
        from orangecontrib.AAIT.utils import MetManagement



    change_owcorpus.replace_owcorpus_file()
# from Orange.widgets.orangecontrib.AAIT.utils.tools import first_time_check # ignore pyflakes alert
# from orangecontrib.AAIT.utils.tools import first_time_check # ignore pyflakes alert
    def remove_temp_file():
        """
        remove file which allows a message box to be displayed “launch of orange software”
        """
        temp_folder = tempfile.gettempdir()
        temp_file_path = os.path.join(temp_folder, 'orange_lance.txt')
        if os.path.exists(temp_file_path):
            try:
                os.remove(temp_file_path)
            except Exception as e:
                print(f"Erreur lors de la suppression du fichier temporaire : {e}")



    def find_widget_discovery_objects():
        """
        return a list of WidgetDiscovery isntance based on garbage collector parsing
        """
        gc.collect()
        objects = gc.get_objects()
        try:
            widget_discoveries = [obj for obj in objects if isinstance(obj, WidgetDiscovery)]
        except Exception as e:
            print(e)
        return widget_discoveries

    def import_proprietary_categories():
        # Utilisation de la fonction pour obtenir les objets
        widget_discoveries = find_widget_discovery_objects()
        the_discovery = None
        for widget in widget_discoveries:
            the_discovery = widget
        if the_discovery is None:
            return
        pkgs = MetManagement.get_category_extension_to_load()
        if len(pkgs)==0:
            return
        sys.path.append(MetManagement.get_widget_extension_path())
        dist = get_distribution("Orange3")
        for pkg in pkgs:
            the_discovery.handle_category(category_from_package_globals(pkg))
        for pkg in pkgs:
            the_discovery.process_category_package(pkg, distribution=dist)

    def duplicate_widget_if_needed_exept_POW_file():
        path_to_check=os.path.dirname(__file__)+"/widgets"
        path_to_check=path_to_check.replace("\\","/")
        files_py = [f for f in os.listdir(path_to_check) if f.endswith('.py')]


        files_py_duplicated=[]
        files_py_not_duplicated=[]
        for element in files_py:
            if element[:2]=="__":
                files_py_duplicated.append(element)
                continue
            files_py_not_duplicated.append(element)
        for element in files_py_duplicated:
            if element=="__init__.py":
                continue
            file_origine=(element.split("__" )[-1])
            if not os.path.isfile(path_to_check+"/"+file_origine):
                try:
                    os.remove(path_to_check+"/"+element)
                    continue
                except Exception as e:
                    print(f"Error : {e}")
            if MetManagement.get_size(path_to_check+"/"+file_origine)!=MetManagement.get_size(path_to_check+"/"+element):
                # Paths to the two files
                file_1 = path_to_check+"/"+file_origine
                file_2 = path_to_check+"/"+element

                try:
                    # Read the content of the first file
                    with open(file_1, 'r') as source_file:
                        content = source_file.read()

                    # Write the content into the second file
                    with open(file_2, 'w') as target_file:
                        target_file.write(content)

                    print(f"The file '{file_2}' has been overwritten with the content of '{file_1}'.")
                    continue
                except FileNotFoundError as e:
                    print(f"Error: File not found. {e}")
                except PermissionError as e:
                    print(f"Error: Permission denied. {e}")
                except Exception as e:
                    print(f"An unexpected error occurred: {e}")


        list_file_to_created=[]
        list_file_to_duplicate=[]
        for extention in MetManagement.get_category_extension_to_load():
            dev=False
            prefix = ""
            # extention_a_jouter=extention
            if len(extention) > 4:
                if extention[:3]=="dev":
                    prefix="__dev"
                    dev=True
                    # case not dev  and dev module required
                    if "site-packages/Orange/widgets" not in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
                        continue
            # case dev  and not dev module required
            if dev == False:
                if "site-packages/Orange/widgets"  in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
                    continue
            prefix+="__"+extention+"__"
            dir_to_check=MetManagement.get_widget_extension_path()+extention+"/"
            files_py = [f for f in os.listdir(dir_to_check) if f.endswith('.py')]
            for ffile in files_py:
                if ffile=="__init__.py":
                    continue
                if ffile[:3]!="POW":
                    if ffile in files_py_not_duplicated:
                        list_file_to_created.append(prefix+ffile)
                        list_file_to_duplicate.append(ffile)
                    continue


        # remove unusuable files
        for element in files_py_duplicated:
            if element =="__init__.py":
                continue
            if element in list_file_to_created:
                continue
            try:
                print("remove ",path_to_check + "/" + element)
                os.remove(path_to_check + "/" + element)
            except Exception as e:
                print(f"Error : {e}")

        for i in range(len(list_file_to_created)):
            #print(list_file_to_created[i]," ",list_file_to_duplicate[i])
            file_1 = path_to_check + "/" + list_file_to_duplicate[i]
            file_2 = path_to_check + "/" + list_file_to_created[i]
            if os.path.isfile(file_2):
                #print(file_2 ,"existe")
                continue

            try:
                # Read the content of the first file
                with open(file_1, 'r') as source_file:
                    content = source_file.read()

                # Write the content into the second file
                with open(file_2, 'w') as target_file:
                    target_file.write(content)

                print(f"The file '{file_2}' has been overwritten with the content of '{file_1}'.")
                continue
            except FileNotFoundError as e:
                print(f"Error: File not found. {e}")
            except PermissionError as e:
                print(f"Error: Permission denied. {e}")
            except Exception as e:
                print(f"An unexpected error occurred: {e}")

    def duplicate_POW_file():
        path_to_check=os.path.dirname(__file__)+"/widgets"
        path_to_check=path_to_check.replace("\\","/")
        files_py = [f for f in os.listdir(path_to_check) if f.endswith('.py') and f.startswith('POW')]
        # update pow file if necessary
        original_element="POW_Wfactory.py"
        original_pow_file=path_to_check+"/"+original_element
        if False==os.path.isfile(original_pow_file):
            print(original_element +"does not exist")
            return
        files_py_duplicated=[]
        for element in files_py:
            if element!=original_element:
                files_py_duplicated.append(element)
            if MetManagement.get_size(original_pow_file) != MetManagement.get_size(
                    path_to_check + "/" + element):
                # Paths to the two files
                file_1 = original_pow_file
                file_2 = path_to_check + "/" + element
                try:
                    # Read the content of the first file
                    with open(file_1, 'r') as source_file:
                        content = source_file.read()

                    # Write the content into the second file
                    with open(file_2, 'w') as target_file:
                        target_file.write(content)

                    print(f"The file '{file_2}' has been overwritten with the content of '{file_1}'.")
                    continue
                except FileNotFoundError as e:
                    print(f"Error: File not found. {e}")
                except PermissionError as e:
                    print(f"Error: Permission denied. {e}")
                except Exception as e:
                    print(f"An unexpected error occurred: {e}")
        list_file_to_created = []
        list_file_to_duplicate = []

        for extention in MetManagement.get_category_extension_to_load():
            dev = False
            suffix = ""
            if len(extention) > 4:
                if extention[:3] == "dev":
                    suffix = "__dev"
                    dev = True
                    # case not dev  and dev module required
                    if "site-packages/Orange/widgets" not in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
                        continue
            # case dev  and not dev module required
            if dev == False:
                if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
                    continue
            suffix+="__"+extention+"__"
            dir_to_check=MetManagement.get_widget_extension_path()+extention+"/"
            files_py = [f for f in os.listdir(dir_to_check) if f.endswith('.py') and f.startswith('POW')]
            for ffile in files_py:
                list_file_to_created.append(ffile[:-3]+suffix+".py")
                list_file_to_duplicate.append(original_element)
        # remove unusuable files
        for element in files_py_duplicated:
            if element in list_file_to_created:
                continue
            try:
                print("remove ",path_to_check + "/" + element)
                os.remove(path_to_check + "/" + element)
            except Exception as e:
                print(f"Error : {e}")
        for i in range(len(list_file_to_created)):
            #print(list_file_to_created[i]," ",list_file_to_duplicate[i])
            file_1 = path_to_check + "/" + list_file_to_duplicate[i]
            file_2 = path_to_check + "/" + list_file_to_created[i]
            if os.path.isfile(file_2):
                #print(file_2 ,"existe")
                continue

            try:
                # Read the content of the first file
                with open(file_1, 'r') as source_file:
                    content = source_file.read()

                # Write the content into the second file
                with open(file_2, 'w') as target_file:
                    target_file.write(content)

                print(f"The file '{file_2}' has been overwritten with the content of '{file_1}'.")
                continue
            except FileNotFoundError as e:
                print(f"Error: File not found. {e}")
            except PermissionError as e:
                print(f"Error: Permission denied. {e}")
            except Exception as e:
                print(f"An unexpected error occurred: {e}")

    # remove the file in temp folder
    remove_temp_file()
    import_proprietary_categories()
    duplicate_widget_if_needed_exept_POW_file()

    duplicate_POW_file()



from AnyQt import QtWidgets, QtGui, QtCore

def force_native_light_mode():
    # 1. Get the current application instance
    app = QtWidgets.QApplication.instance()
    if not app:
        return

    # 2. Force the style back to 'Windows' or 'WindowsVista'
    # This prevents Orange from using any custom Dark-specific styles
    app.setStyle(QtWidgets.QStyleFactory.create("WindowsVista"))

    # 3. Explicitly tell Qt to use the Light Palette
    # We fetch the 'Standard' palette which defaults to Light colors
    light_palette = QtGui.QPalette()

    # Manually re-assert the Light Mode colors to override OS injection
    light_palette.setColor(QtGui.QPalette.Window, QtGui.QColor(240, 240, 240))
    light_palette.setColor(QtGui.QPalette.WindowText, QtCore.Qt.black)
    light_palette.setColor(QtGui.QPalette.Base, QtCore.Qt.white)
    light_palette.setColor(QtGui.QPalette.AlternateBase, QtGui.QColor(233, 233, 233))
    light_palette.setColor(QtGui.QPalette.ToolTipBase, QtCore.Qt.white)
    light_palette.setColor(QtGui.QPalette.ToolTipText, QtCore.Qt.black)
    light_palette.setColor(QtGui.QPalette.Text, QtCore.Qt.black)
    light_palette.setColor(QtGui.QPalette.Button, QtGui.QColor(240, 240, 240))
    light_palette.setColor(QtGui.QPalette.ButtonText, QtCore.Qt.black)
    light_palette.setColor(QtGui.QPalette.BrightText, QtCore.Qt.red)
    light_palette.setColor(QtGui.QPalette.Link, QtGui.QColor(0, 0, 255))
    light_palette.setColor(QtGui.QPalette.Highlight, QtGui.QColor(0, 120, 215))
    light_palette.setColor(QtGui.QPalette.HighlightedText, QtCore.Qt.white)

    app.setPalette(light_palette)

force_native_light_mode()


import os
import sys

import Orange.data
from Orange.data import Table, Domain, ContinuousVariable
from AnyQt.QtWidgets import QApplication
from Orange.widgets import widget
from Orange.widgets.utils.signals import Input, Output

from sentence_transformers import CrossEncoder

if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils import thread_management
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.utils import thread_management
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file



@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWReranking(widget.OWWidget):
    name = "Reranking"
    description = "Apply a reranker to the column 'content' of your input data, in regards to a column 'request' in your request data."
    category = "AAIT - TOOLBOX"
    icon = "icons/owreranking.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owreranking.svg"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owreranking.ui")
    want_control_area = False
    priority = 1212

    class Inputs:
        data = Input("Data", Orange.data.Table)
        request = Input("Request", Orange.data.Table)
        model_path = Input("Model", str, auto_summary=False)

    class Outputs:
        data = Output("Data", Orange.data.Table)

    @Inputs.data
    def set_data(self, in_data):
        self.data = in_data
        # if self.autorun:
        #     self.run()

    @Inputs.request
    def set_request(self, in_request):
        self.data_request = in_request
        # if self.autorun:
        #     self.run()

    @Inputs.model_path
    def set_model_path(self, in_model_path):
        self.model_path = in_model_path
        # if self.autorun:
        #     self.run()

    def handleNewSignals(self):
        self.run()

    def __init__(self):
        super().__init__()
        # Qt Management
        self.setFixedWidth(470)
        self.setFixedHeight(300)
        uic.loadUi(self.gui, self)

        # Data Management
        self.data = None
        self.data_request = None
        self.model_path = None
        self.model = None
        self.thread = None
        self.autorun = True
        self.result = None
        self.post_initialized()


    def run(self):
        self.error("")
        self.warning("")

        # if thread is running quit
        if self.thread is not None:
            self.thread.safe_quit()

        if self.data is None:
            self.Outputs.data.send(None)
            return

        if self.data_request is None:
            self.Outputs.data.send(None)
            return

        if self.model_path is None:
            self.Outputs.data.send(None)
            return

        self.load_model()
        if self.model is None:
            self.Outputs.data.send(None)
            return

        # Verification of in_data
        if not "content" in self.data.domain or type(self.data.domain["content"]).__name__ != "StringVariable":
            self.error('You need a "content" column as a text variable in your input data')
            return

        if not "request" in self.data_request.domain or type(self.data_request.domain["request"]).__name__ != "StringVariable":
            self.error('You need a "request" column as a text variable in your input request')
            return


        request = self.data_request[0]["request"].value

        # Start progress bar
        self.progressBarInit()

        # Connect and start thread : main function, progress, result and finish
        # --> progress is used in the main function to track progress (with a callback)
        # --> result is used to collect the result from main function
        # --> finish is just an empty signal to indicate that the thread is finished
        self.thread = thread_management.Thread(self.rerank_table, self.data, request, self.model)
        self.thread.progress.connect(self.handle_progress)
        self.thread.result.connect(self.handle_result)
        self.thread.finish.connect(self.handle_finish)
        self.thread.start()

    def handle_progress(self, value: float) -> None:
        self.progressBarSet(value)

    def handle_result(self, result):
        try:
            self.result = result
            self.Outputs.data.send(result)
            # self.data = None
            # self.data_request = None
        except Exception as e:
            print("An error occurred when sending out_data:", e)
            self.Outputs.data.send(None)
            return

    def handle_finish(self):
        print("Reranking finished")
        self.progressBarFinished()

    def post_initialized(self):
        pass


    def load_model(self):
        self.error("")
        try:
            self.model = CrossEncoder(self.model_path)
        except Exception as e:
            self.error(f"An error occured when trying to load the model: {e}")
            self.model = None


    def rerank_table(self, table, request, model, progress_callback=None, argself=None):
        # Copy of input data
        data = table.copy()
        attr_dom = list(data.domain.attributes)
        metas_dom = list(data.domain.metas)
        class_dom = list(data.domain.class_vars)

        # Get the "content" but keep the entire row
        documents = [(row["content"].value, row) for row in data]
        # Create request - content pairs for scoring
        pairs = [(request, content) for content, _ in documents]
        scores = model.predict(pairs)

        # Combine scores with original rows
        scored_rows = list(zip(scores, documents))
        sorted_rows = sorted(scored_rows, key=lambda x: x[0], reverse=True)

        # Iterate on the data Table
        rows = []
        for i, (score, (_, row)) in enumerate(sorted_rows):
            # Get the rest of the data
            features = [row[x] for x in attr_dom]
            targets = [row[y] for y in class_dom]
            metas = list(row.metas)
            new_row = features + targets + metas + [score]
            rows.append(new_row)
            if progress_callback is not None:
                progress_value = float(100 * (i + 1) / len(sorted_rows))
                progress_callback(progress_value)
            if argself is not None:
                if argself.stop:
                    break

        # Create new Domain for new columns
        score_var = [ContinuousVariable("Score (reranking)")]
        domain = Domain(attributes=attr_dom, metas=metas_dom + score_var, class_vars=class_dom)

        # Create and return table
        out_data = Table.from_list(domain=domain, rows=rows)
        return out_data


if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWReranking()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()



from orangecanvas.localization import Translator  # pylint: disable=wrong-import-order
_tr = Translator("Orange", "biolab.si", "Orange")
del Translator
import concurrent.futures
from dataclasses import dataclass
from typing import (
    Optional, Union, Sequence, List, TypedDict, Tuple, Any, Container
)

from scipy.sparse import issparse

from AnyQt.QtWidgets import (
    QTableView, QHeaderView, QApplication, QStyle, QStyleOptionHeader,
    QStyleOptionViewItem
)
from AnyQt.QtGui import QColor, QClipboard, QPainter
from AnyQt.QtCore import (
    Qt, QSize, QMetaObject, QItemSelectionModel, QModelIndex, QRect
)
from AnyQt.QtCore import Slot

from orangewidget.gui import OrangeUserRole

import Orange.data
from Orange.data.table import Table
from Orange.data.sql.table import SqlTable

from Orange.widgets import gui
from Orange.widgets.data.utils.models import RichTableModel, TableSliceProxy
from Orange.widgets.utils.itemdelegates import TableDataDelegate
from Orange.widgets.utils.tableview import table_selection_to_mime_data
from Orange.widgets.utils.widgetpreview import WidgetPreview
from Orange.widgets.widget import OWWidget, Input, Output, Msg
from Orange.widgets.utils.annotated_data import (create_annotated_table,
                                                 ANNOTATED_DATA_SIGNAL_NAME)
from Orange.widgets.utils.itemmodels import TableModel
from Orange.widgets.utils.state_summary import format_summary_details
from Orange.widgets.utils import disconnected
from Orange.widgets.utils.headerview import HeaderView
from Orange.widgets.data.utils.tableview import RichTableView
from Orange.widgets.data.utils import tablesummary as tsummary
from Orange.widgets.settings import Setting
import os
if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils import widget_positioning
else:
    from orangecontrib.AAIT.utils import widget_positioning


SubsetRole = next(OrangeUserRole)


class HeaderViewWithSubsetIndicator(HeaderView):
    _IndicatorChar = "\N{BULLET}"

    def paintSection(
            self, painter: QPainter, rect: QRect, logicalIndex: int
    ) -> None:
        opt = QStyleOptionHeader()
        self.initStyleOption(opt)
        self.initStyleOptionForIndex(opt, logicalIndex)
        model = self.model()
        if model is None:
            return  # pragma: no cover
        opt.rect = rect
        issubset = model.headerData(logicalIndex, Qt.Vertical, SubsetRole)
        style = self.style()
        # draw background
        style.drawControl(QStyle.CE_HeaderSection, opt, painter, self)
        indicator_rect = QRect(rect)
        text_rect = QRect(rect)
        indicator_width = opt.fontMetrics.horizontalAdvance(
            self._IndicatorChar + " "
        )
        indicator_rect.setWidth(indicator_width)
        text_rect.setLeft(indicator_width)
        if issubset:
            optindicator = QStyleOptionHeader(opt)
            optindicator.rect = indicator_rect
            optindicator.textAlignment = Qt.AlignCenter
            optindicator.text = self._IndicatorChar
            # draw subset indicator
            style.drawControl(QStyle.CE_HeaderLabel, optindicator, painter, self)
        opt.rect = text_rect
        # draw section label
        style.drawControl(QStyle.CE_HeaderLabel, opt, painter, self)

    def sectionSizeFromContents(self, logicalIndex: int) -> QSize:
        opt = QStyleOptionHeader()
        self.initStyleOption(opt)
        super().initStyleOptionForIndex(opt, logicalIndex)
        opt.text = self._IndicatorChar + " " + opt.text
        return self.style().sizeFromContents(QStyle.CT_HeaderSection, opt, QSize(), self)


class DataTableView(gui.HScrollStepMixin, RichTableView):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        vheader = HeaderViewWithSubsetIndicator(
            Qt.Vertical, self, highlightSections=True
        )
        vheader.setSectionsClickable(True)
        self.setVerticalHeader(vheader)


class _TableDataDelegate(TableDataDelegate):
    DefaultRoles = TableDataDelegate.DefaultRoles + (SubsetRole,)


class SubsetTableDataDelegate(_TableDataDelegate):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subset_opacity = 0.5

    def paint(
            self, painter: QPainter, option: QStyleOptionViewItem,
            index: QModelIndex
    ) -> None:
        issubset = self.cachedData(index, SubsetRole)
        opacity = painter.opacity()
        if not issubset:
            painter.setOpacity(self.subset_opacity)
        super().paint(painter, option, index)
        if not issubset:
            painter.setOpacity(opacity)


class TableBarItemDelegate(SubsetTableDataDelegate, gui.TableBarItem,
                           _TableDataDelegate):
    pass


class _TableModel(RichTableModel):
    SubsetRole = SubsetRole

    def __init__(self, *args, subsets=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._subset = subsets or set()

    def setSubsetRowIds(self, subsetids: Container[int]):
        self._subset = subsetids
        if self.rowCount():
            self.headerDataChanged.emit(Qt.Vertical, 0, self.rowCount() - 1)
            self.dataChanged.emit(
                self.index(0, 0),
                self.index(self.rowCount() - 1, self.columnCount() - 1),
                [SubsetRole],
            )

    def _is_subset(self, row):
        row = self.mapToSourceRows(row)
        try:
            id_ = self.source.ids[row]
        except (IndexError, AttributeError):  # pragma: no cover
            return False
        return int(id_) in self._subset

    def data(self, index: QModelIndex, role=Qt.DisplayRole) -> Any:
        if role == _TableModel.SubsetRole:
            return self._is_subset(index.row())
        return super().data(index, role)

    def headerData(self, section, orientation, role):
        if orientation == Qt.Vertical and role == _TableModel.SubsetRole:
            return self._is_subset(section)
        return super().headerData(section, orientation, role)


@dataclass
class InputData:
    table: Table
    summary: Union[tsummary.Summary, tsummary.ApproxSummary]
    model: TableModel


class _Selection(TypedDict):
    rows: Tuple[int]
    columns: Tuple[int]


_Sorting = List[Tuple[str, int]]


class OWTable(OWWidget):
    name = "Autoshow Data Table"
    description = "View the dataset in a spreadsheet."
    icon = "icons/Table.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/Table.svg"
    category = "AAIT - TOOLBOX"
    priority = 50
    keywords = "data table, view"

    class Inputs:
        data = Input("Data", Table, default=True)
        data_subset = Input("Data Subset", Table)
        input_autoshow = Input("AutoShowConfiguration", str, auto_summary=False)

    class Outputs:
        selected_data = Output("Selected Data", Table, default=True)
        annotated_data = Output(ANNOTATED_DATA_SIGNAL_NAME, Table)

    class Warning(OWWidget.Warning):
        missing_sort_columns = Msg(
            ( "Cannot restore sorting.\n" + "Missing columns in input table: {}")
        )
        non_sortable_input = Msg(
            ( "Cannot restore sorting.\n" + "Input table cannot be sorted due to implementation constraints.")
        )

    str_WidgetPositionning: str=Setting("None")
    buttons_area_orientation = Qt.Vertical

    show_distributions = Setting(False)
    show_attribute_labels = Setting(True)
    select_rows = Setting(True)
    auto_commit = Setting(True)

    color_by_class = Setting(True)
    stored_selection: _Selection = Setting(
        {"rows": [], "columns": []}, schema_only=True
    )
    stored_sort: _Sorting = Setting(
        [], schema_only=True
    )
    settings_version = 1

    def __init__(self):
        super().__init__()
        self.input: Optional[InputData] = None
        self._subset_ids: Optional[set] = None
        self.__pending_selection: Optional[_Selection] = self.stored_selection
        self.__pending_sort: Optional[_Sorting] = self.stored_sort
        self.__have_new_data = False
        self.__have_new_subset = False
        self.dist_color = QColor(220, 220, 220, 255)

        info_box = gui.vBox(self.controlArea, "Info")
        self.info_text = gui.widgetLabel(info_box)

        box = gui.vBox(self.controlArea,  "Variables")
        self.c_show_attribute_labels = gui.checkBox(
            box, self, "show_attribute_labels",
            "Show variable labels (if present)",
            callback=self._update_variable_labels)

        gui.checkBox(box, self, "show_distributions",
                     'Visualize numeric values',
                     callback=self._on_distribution_color_changed)
        gui.checkBox(box, self, "color_by_class", 'Color by instance classes',
                     callback=self._on_distribution_color_changed)

        box = gui.vBox(self.controlArea,  "Selection")

        gui.checkBox(box, self, "select_rows", "Select full rows",
                     callback=self._on_select_rows_changed)

        gui.rubber(self.controlArea)

        gui.button(self.buttonsArea, self,  "Restore Original Order",
                   callback=self.restore_order,
                   tooltip="Show rows in the original order",
                   autoDefault=False,
                   attribute=Qt.WA_LayoutUsesWidgetRect)
        gui.auto_send(self.buttonsArea, self, "auto_commit")

        view = DataTableView(sortingEnabled=True)
        view.setItemDelegate(SubsetTableDataDelegate(view))
        view.selectionFinished.connect(self.update_selection)

        if self.select_rows:
            view.setSelectionBehavior(QTableView.SelectRows)

        header = view.horizontalHeader()
        header.setSectionsMovable(True)
        header.setSectionsClickable(True)
        header.setSortIndicatorShown(True)
        header.setSortIndicator(-1, Qt.AscendingOrder)
        header.sortIndicatorChanged.connect(
            self._on_sort_indicator_changed, Qt.UniqueConnection
        )

        self.view = view
        self.mainArea.layout().addWidget(self.view)
        self._update_input_summary()
        widget_positioning.show_and_adjust_at_opening(self,str(self.str_WidgetPositionning))


    def copy_to_clipboard(self):
        self.copy()

    def sizeHint(self):
        return QSize(800, 500)

    @Inputs.data
    def set_dataset(self, data: Optional[Table]):
        """Set the input dataset."""
        if data is not None:
            summary = tsummary.table_summary(data)
            self.input = InputData(
                table=data,
                summary=summary,
                model=_TableModel(data)
            )
            if isinstance(summary.len, concurrent.futures.Future):
                def update(_):
                    QMetaObject.invokeMethod(
                        self, "_update_info", Qt.QueuedConnection)
                summary.len.add_done_callback(update)
        else:
            self.input = None
        self.__have_new_data = True

    @Inputs.data_subset
    def set_subset_dataset(self, subset: Optional[Table]):
        """Set the data subset"""
        if subset is not None and not isinstance(subset, SqlTable):
            ids = set(subset.ids)
        else:
            ids = None
        self._subset_ids = ids
        self.__have_new_subset = True

    @Inputs.input_autoshow
    def set_input_autoshow(self, le_str):
        if le_str is not None:
            self.str_WidgetPositionning=str(le_str)


    def handleNewSignals(self):
        super().handleNewSignals()
        self.Warning.non_sortable_input.clear()
        self.Warning.missing_sort_columns.clear()
        data: Optional[Table] = self.input.table if self.input else None
        model = self.input.model if self.input else None

        if self.__have_new_data:
            self._setup_table_view()
            self._update_input_summary()

            if data is not None and self.__pending_sort is not None:
                self.__restore_sort()

            if data is not None and self.__pending_selection is not None:
                selection = self.__pending_selection
                self.__pending_selection = None
                rows = selection["rows"]
                columns = selection["columns"]
                self.set_selection(rows, columns)

        if self.__have_new_subset and model is not None:
            model.setSubsetRowIds(self._subset_ids or set())
            self.__have_new_subset = False

        self._setup_view_delegate()

        if self.__have_new_data:
            self.commit.now()
            self.__have_new_data = False

    def _setup_table_view(self):
        """Setup the view with current input data."""
        if self.input is None:
            self.view.setModel(None)
            return

        datamodel = self.input.model
        datamodel.setSubsetRowIds(self._subset_ids or set())

        view = self.view
        data = self.input.table
        rowcount = data.approx_len()
        view.setModel(datamodel)

        vheader = view.verticalHeader()
        option = view.viewOptions()
        size = view.style().sizeFromContents(
            QStyle.CT_ItemViewItem, option,
            QSize(20, 20), view)

        vheader.setDefaultSectionSize(size.height() + 2)
        vheader.setMinimumSectionSize(5)
        vheader.setSectionResizeMode(QHeaderView.Fixed)

        # Limit the number of rows displayed in the QTableView
        # (workaround for QTBUG-18490 / QTBUG-28631)
        maxrows = (2 ** 31 - 1) // (vheader.defaultSectionSize() + 2)
        if rowcount > maxrows:
            sliceproxy = TableSliceProxy(
                parent=view, rowSlice=slice(0, maxrows))
            sliceproxy.setSourceModel(datamodel)
            # First reset the view (without this the header view retains
            # it's state - at this point invalid/broken)
            view.setModel(None)
            view.setModel(sliceproxy)

        assert view.model().rowCount() <= maxrows
        assert vheader.sectionSize(0) > 1 or datamodel.rowCount() == 0

        self._setup_view_delegate()
        # update the header (attribute names)
        self._update_variable_labels()

    def _update_input_summary(self):
        def format_summary(summary):
            if isinstance(summary, tsummary.ApproxSummary):
                length = summary.len.result() if summary.len.done() else \
                    summary.approx_len
            elif isinstance(summary, tsummary.Summary):
                length = summary.len
            return length

        summary, details = self.info.NoInput, ""
        if self.input:
            summary = format_summary(self.input.summary)
            details = format_summary_details(self.input.table)
        self.info.set_input_summary(summary, details)

        if self.input is None:
            summary = [ "No data."]
        else:
            summary = tsummary.format_summary(self.input.summary)
        self.info_text.setText("\n".join(summary))

    def _update_variable_labels(self):
        """Update the variable labels visibility for current view."""
        if self.input is None:
            return
        model = self.input.model
        if self.show_attribute_labels:
            model.setRichHeaderFlags(
                RichTableModel.Labels | RichTableModel.Name
            )
        else:
            model.setRichHeaderFlags(RichTableModel.Name)

    def _on_distribution_color_changed(self):
        if self.input is None:
            return  # pragma: no cover
        self._setup_view_delegate()

    def _setup_view_delegate(self):
        if self.input is None:
            return
        model = self.input.model
        data = model.source
        class_var = data.domain.class_var
        if self.color_by_class and class_var and class_var.is_discrete:
            color_schema = [QColor(*c) for c in class_var.colors]
        else:
            color_schema = None
        if self.show_distributions:
            delegate = TableBarItemDelegate(
                self.view, color=self.dist_color, color_schema=color_schema
            )
        else:
            delegate = SubsetTableDataDelegate(self.view)
        delegate.subset_opacity = 0.5 if self._subset_ids is not None else 1.0
        self.view.setItemDelegate(delegate)

    def _on_select_rows_changed(self):
        if self.input is None:
            return
        selection_model = self.view.selectionModel()
        selection_model.setSelectBlocks(not self.select_rows)
        if self.select_rows:
            self.view.setSelectionBehavior(QTableView.SelectRows)
            # Expand the current selection to full row selection.
            selection_model.select(
                selection_model.selection(),
                QItemSelectionModel.Select | QItemSelectionModel.Rows
            )
        else:
            self.view.setSelectionBehavior(QTableView.SelectItems)

    def restore_order(self):
        """Restore the original data order of the current view."""
        self.view.sortByColumn(-1, Qt.AscendingOrder)
        self.stored_sort = []
        self.Warning.missing_sort_columns.clear()

    @Slot()
    def _update_info(self):
        self._update_input_summary()

    def _on_sort_indicator_changed(self, index: int, order: Qt.SortOrder) -> None:
        if index == -1:
            self.stored_sort = []
        elif self.input is not None:
            model = self.input.model
            coldesc = model.columns[index]
            colid = self.__encode_column_id(coldesc)
            order = -1 if order == Qt.DescendingOrder else 1
            # Drop any previously applied sort on this column
            self.stored_sort = [(n, d) for n, d in self.stored_sort
                                if n != colid]
            self.stored_sort.append((colid, order))
        self.update_selection()
        self.Warning.missing_sort_columns.clear()

    def set_sort_columns(self, sorting: List[Tuple[str, int]]):
        """
        Set the model sorting parameters.

        Parameters
        ----------
        sorting: List[Tuple[str, int]]
            For each (name: str, inc: int) tuple where `name` is the column
            name and `inc` is 1 for increasing order and -1 for decreasing
            order, the model is sorted by that column.
        """
        if self.input is None:
            return  # pragma: no cover
        self.stored_sort = []
        # Map header ids (names) to column indices
        columns = {id_: i for i, id_ in enumerate(self.__header_ids())}
        # Suppress the _on_sort_indicator_changed -> commit calls
        with disconnected(self.view.horizontalHeader().sortIndicatorChanged,
                          self._on_sort_indicator_changed, Qt.UniqueConnection):
            for colid, order in sorting:
                if colid in columns:
                    self.view.sortByColumn(
                        columns[colid],
                        Qt.AscendingOrder if order == 1 else Qt.DescendingOrder
                    )
                self.stored_sort.append((colid, order))

    def __restore_sort(self) -> None:
        assert self.input is not None
        sort = self.__pending_sort
        self.__pending_sort = None
        if sort is None:
            return  # pragma: no cover
        if not self.view.isSortingEnabled() and sort:
            self.Warning.non_sortable_input()
            self.Warning.missing_sort_columns.clear()
            return
        # Map header ids (names) to column indices
        columns = {id_: i for i, id_ in enumerate(self.__header_ids())}
        missing_columns = []
        sort_ = []
        for colid, order in sort:
            if colid in columns:
                sort_.append((colid, order))
            else:
                missing_columns.append(self.__decode_column_id(colid))
        self.set_sort_columns(sort_)
        if missing_columns:
            self.Warning.missing_sort_columns(", ".join(missing_columns))

    @staticmethod
    def __encode_column_id(
            coldesc: Union[TableModel.Column, TableModel.Basket]
    ) -> str:
        def escape(s: str) -> str:  # escape possible leading slash
            if s.startswith("\\"):
                return "\\" + s
            return s
        if isinstance(coldesc, TableModel.Column):
            return escape(coldesc.var.name)
        else:
            lookup = ("TARGET", "META", "FEATURES",)
            return f"\\BASKET({lookup[coldesc.role]})"

    @staticmethod
    def __decode_column_id(cid: str) -> str:
        if cid.startswith("\\"):
            return cid[1:]
        return cid

    def __header_ids(self) -> List[str]:
        if self.input is None:
            return []
        return [self.__encode_column_id(c) for c in self.input.model.columns]

    def update_selection(self, *_):
        self.commit.deferred()

    def set_selection(self, rows: Sequence[int], columns: Sequence[int]) -> None:
        """
        Set the selected `rows` and `columns`.

        `rows` are indices into underlying :class:`Table`
        """
        self.view.setBlockSelection(rows, columns)

    def get_selection(self):
        """
        Return the selected row and column indices of the selection in view.
        """
        return self.view.blockSelection()

    @gui.deferred
    def commit(self):
        """
        Commit/send the current selected row/column selection.
        """
        selected_data = table = rowsel = None
        if self.input is not None:
            model = self.input.model
            table = self.input.table

            # Selections of individual instances are not implemented
            # for SqlTables
            if isinstance(table, SqlTable):
                self.Outputs.selected_data.send(selected_data)
                self.Outputs.annotated_data.send(None)
                return

            rowsel, colsel = self.get_selection()
            self.stored_selection = {"rows": rowsel, "columns": colsel}

            domain = table.domain

            if len(colsel) < len(domain.variables) + len(domain.metas):
                # only a subset of the columns is selected
                allvars = domain.class_vars + domain.metas + domain.attributes
                columns = [(c, model.headerData(c, Qt.Horizontal,
                                                TableModel.DomainRole))
                           for c in colsel]
                assert all(role is not None for _, role in columns)

                def select_vars(role):
                    """select variables for role (TableModel.DomainRole)"""
                    return [allvars[c] for c, r in columns if r == role]

                attrs = select_vars(TableModel.Attribute)
                if attrs and issparse(table.X):
                    # for sparse data you can only select all attributes
                    attrs = table.domain.attributes
                class_vars = select_vars(TableModel.ClassVar)
                metas = select_vars(TableModel.Meta)
                domain = Orange.data.Domain(attrs, class_vars, metas)

            sortsection = self.view.horizontalHeader().sortIndicatorSection()
            if rowsel:
                selected_data = table.from_table(domain, table, rowsel)
            elif sortsection != -1:
                # Send sorted data
                permutation = model.mapToSourceRows(...)
                selected_data = table.from_table(table.domain, table, permutation)
            else:
                # Send all data by default
                selected_data = table

        self.Outputs.selected_data.send(selected_data)
        self.Outputs.annotated_data.send(create_annotated_table(table, rowsel))

    def copy(self):
        """
        Copy current table selection to the clipboard.
        """
        if self.input is not None:
            mime = table_selection_to_mime_data(self.view)
            QApplication.clipboard().setMimeData(
                mime, QClipboard.Clipboard
            )

    def send_report(self):
        if self.input is None:
            return
        model = self.input.model
        self.report_data_brief(model.source)
        self.report_table(self.view)


if __name__ == "__main__":  # pragma: no cover
    WidgetPreview(OWTable).run(
        input_data=Table("iris"),
    )



import os
import re
import io
import sys
import code
import subprocess
import contextlib
import traceback


from Orange.data import Table, Domain, StringVariable
from AnyQt.QtWidgets import QApplication, QCheckBox, QPushButton
from AnyQt import QtWidgets, QtGui
from Orange.widgets.settings import Setting


if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
    from Orange.widgets.orangecontrib.AAIT.utils import thread_management, base_widget
else:
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
    from orangecontrib.AAIT.utils import thread_management, base_widget


@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWExecuteScript(base_widget.BaseListWidget):
    name = "Execute Script"
    description = "Locally execute the scripts contained in the column 'Script'."
    category = "AAIT - TOOLBOX"
    icon = "icons/owexecutescript.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owexecutescript.svg"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owexecutescript_TEST.ui")
    want_control_area = False
    priority = 1060

    # Settings
    interactive = Setting(False)

    # class Inputs:
    #     data = Input("Data", Orange.data.Table)
    #
    # class Outputs:
    #     data = Output("Data", Orange.data.Table)
    #
    # @Inputs.data
    # def set_data(self, in_data):
    #     self.data = in_data
    #     if self.data:
    #         self.var_selector.add_variables(self.data.domain)
    #         self.var_selector.select_variable_by_name(self.selected_column_name)
    #     if self.autorun:
    #         self.run()

    def __init__(self):
        super().__init__()
        # Qt Management
        self.setFixedWidth(470)
        self.setFixedHeight(500)
        # uic.loadUi(self.gui, self)
        # Checkbox
        self.checkbox_mode = self.findChild(QCheckBox, 'checkBox')
        self.checkbox_mode.setChecked(self.interactive)
        self.checkbox_mode.toggled.connect(self.toogle_persistent)
        # Button Show
        self.button_show = self.findChild(QPushButton, 'pushButton')
        self.button_show.clicked.connect(self.show_script)
        # Button Reset
        self.button_reset = self.findChild(QPushButton, "pushButton_2")
        self.button_reset.clicked.connect(self.reset_state)


        # Data Management
        self.data = None
        self.code = None
        self.console = None
        self.script_text = ""
        self.autorun = True
        self.thread = None
        self.result = None
        self.post_initialized()


    def toogle_persistent(self, enabled):
        if enabled:
            self.interactive = True
            if self.console is None:
                self.console = SmartConsole()
        else:
            self.interactive = False

    def reset_state(self):
        self.console = SmartConsole()
        self.script_text = ""

    def show_script(self):
        popup = ScriptPopup(self.script_text, self)
        if hasattr(popup, "exec"):
            popup.exec()
        else:
            popup.exec_()
    # def on_variable_selected(self, var_name):
    #     """Update the selected column when the user clicks an item."""
    #     self.selected_column_name = var_name


    def run(self):
        self.warning("")
        self.error("")

        # If Thread is already running, interrupt it
        if self.thread is not None:
            if self.thread.isRunning():
                self.thread.safe_quit()

        if self.data is None:
            self.Outputs.data.send(None)
            return

        if not self.selected_column_name in self.data.domain:
            self.warning(f'Previously selected column "{self.selected_column_name}" does not exist in your data.')
            return

        if not isinstance(self.data.domain[self.selected_column_name], StringVariable):
            self.error('You must select a text variable.')
            return

        # Start progress bar
        self.progressBarInit()

        # Thread management
        if self.interactive:
            # Interactive mode: Run in main thread for GUI compatibility
            self.result = self.execute_scripts_in_table(self.data, self.selected_column_name)
            self.handle_result(self.result)
            self.handle_finish()
        else:
            self.thread = thread_management.Thread(self.execute_scripts_in_table, self.data, self.selected_column_name)
            self.thread.progress.connect(self.handle_progress)
            self.thread.result.connect(self.handle_result)
            self.thread.finish.connect(self.handle_finish)
            self.thread.start()


    def handle_progress(self, value: float) -> None:
        self.progressBarSet(value)

    def handle_result(self, result):
        try:
            self.result = result
            self.Outputs.data.send(result)
        except Exception as e:
            print("An error occurred when sending out_data:", e)
            self.Outputs.data.send(None)
            return

    def handle_finish(self):
        print("Scripts executed !")
        self.progressBarFinished()


    def post_initialized(self):
        pass


    def execute_scripts_in_table(self, table, selected_var, progress_callback=None, argself=None):
        # Copy of input data
        data = table.copy()
        attr_dom = list(data.domain.attributes)
        metas_dom = list(data.domain.metas)
        class_dom = list(data.domain.class_vars)

        # Iterate on the data Table
        rows = []
        for i, row in enumerate(data):
            # Get the rest of the data
            features = [row[x] for x in attr_dom]
            targets = [row[y] for y in class_dom]
            metas = list(data.metas[i])
            # Execute the script for the given row
            output, error = self.execute_script(row[selected_var].value)
            # Store the output / error
            new_row = features + targets + metas + [output, error]
            rows.append(new_row)
            if progress_callback is not None:
                progress_value = float(100 * (i + 1) / len(data))
                progress_callback(progress_value)
            if argself is not None:
                if argself.stop:
                    break

        # Create new Domain for new columns
        script_dom = [StringVariable("Script output"), StringVariable("Script error")]
        domain = Domain(attributes=attr_dom, metas=metas_dom + script_dom, class_vars=class_dom)
        # Create and return table
        out_data = Table.from_list(domain=domain, rows=rows)
        return out_data


    def execute_script(self, script: str):
        if self.interactive:
            if self.console is None:
                self.console = SmartConsole()

            executable_code = f"exec({script!r})"
            self.script_text += f"\n\n{script}" if self.script_text else script


            stdout, stderr = io.StringIO(), io.StringIO()
            with contextlib.redirect_stdout(stdout), contextlib.redirect_stderr(stderr):
                self.console.last_error = ""
                try:
                    self.console.push(executable_code)
                except Exception:
                    pass  # all errors are captured in showtraceback

            output = stdout.getvalue()
            # Append any exception trace
            error = stderr.getvalue() + self.console.last_error
            return output, error

        else:
            try:
                result = subprocess.run(
                    [sys.executable, "-c", script],
                    capture_output=True,
                    text=True,
                    check=True
                )
                return result.stdout, result.stderr
            except subprocess.CalledProcessError as e:
                return e.stdout, e.stderr


class SmartConsole(code.InteractiveConsole):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.last_error = ""
        self.push("import matplotlib; matplotlib.use('QtAgg')")

    def push(self, line):
        if line.startswith("!pip "):
            cmd = line[1:].split()
            subprocess.check_call([sys.executable, "-m"] + cmd)
            return False
        return super().push(line)

    def showtraceback(self):
        # Capture the exception instead of printing to terminal
        self.last_error = traceback.format_exc()




class PythonHighlighter(QtGui.QSyntaxHighlighter):
    def __init__(self, document):
        super().__init__(document)

        # Styles simples : couleur + gras/italique
        keyword_format = QtGui.QTextCharFormat()
        keyword_format.setForeground(QtGui.QColor("#569CD6"))
        keyword_format.setFontWeight(QtGui.QFont.Bold)

        builtin_format = QtGui.QTextCharFormat()
        builtin_format.setForeground(QtGui.QColor("#4EC9B0"))

        string_format = QtGui.QTextCharFormat()
        string_format.setForeground(QtGui.QColor("#CE9178"))

        comment_format = QtGui.QTextCharFormat()
        comment_format.setForeground(QtGui.QColor("#6A9955"))
        comment_format.setFontItalic(True)

        number_format = QtGui.QTextCharFormat()
        number_format.setForeground(QtGui.QColor("#B5CEA8"))

        # Expressions régulières associées aux styles
        self.rules = []
        keywords = [
            "def", "class", "if", "elif", "else", "while", "for", "try", "except",
            "finally", "with", "as", "return", "import", "from", "pass", "break",
            "continue", "and", "or", "not", "in", "is", "lambda", "yield", "None",
            "True", "False"
        ]
        self.rules += [(r"\b" + kw + r"\b", keyword_format) for kw in keywords]

        builtins = [
            "print", "len", "range", "int", "float", "str", "list", "dict", "set", "tuple"
        ]
        self.rules += [(r"\b" + bi + r"\b", builtin_format) for bi in builtins]

        self.rules.append((r"\".*?\"|'.*?'", string_format))   # chaînes
        self.rules.append((r"#.*", comment_format))            # commentaires
        self.rules.append((r"\b[0-9]+\b", number_format))      # nombres

    def highlightBlock(self, text):
        for pattern, fmt in self.rules:
            for match in re.finditer(pattern, text):
                start, end = match.span()
                self.setFormat(start, end - start, fmt)





# --- Popup window to display code ---
class ScriptPopup(QtWidgets.QDialog):
    def __init__(self, code_text="", parent=None):
        super().__init__(parent)
        self.setWindowTitle("Script Viewer")
        self.resize(600, 400)

        layout = QtWidgets.QVBoxLayout(self)

        self.code_view = QtWidgets.QPlainTextEdit()
        self.code_view.setReadOnly(True)
        self.code_view.setFont(QtGui.QFont("Courier New", 10))
        self.code_view.setPlainText(code_text)

        layout.addWidget(self.code_view)

        # Apply Python syntax highlighting
        self.highlighter = PythonHighlighter(self.code_view.document())



if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWExecuteScript()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()



import os
import sys

from AnyQt.QtWidgets import QApplication, QTreeWidgetItem
from Orange.widgets.settings import Setting
from Orange.widgets import widget
from orangecanvas.application.canvasmain import CanvasMainWindow
from collections import defaultdict

from AnyQt.QtGui import QFont, QColor
from AnyQt.QtCore import Qt

# On prépare le style pour la colonne de droite
italic_font = QFont()
italic_font.setItalic(True)
italic_font.setPointSize(9) # Optionnel: un poil plus petit pour le nom technique
gray_color = QColor("gray")

if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file


@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWAutoShow(widget.OWWidget):
    name = "Auto Show (UNSTABLE)"
    description = "This widget allows you to select which widget you want to automatically open when starting a workflow."
    category = "AAIT - TOOLBOX"
    icon = "icons/autoshow.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/autoshow.svg"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owautoshow.ui")
    want_control_area = False
    priority = 1060

    # Settings
    selected_widgets = Setting([])

    def __init__(self):
        super().__init__()
        # Qt Management
        self.setFixedWidth(470)
        self.setFixedHeight(450)
        uic.loadUi(self.gui, self)

        self.thread = None
        self.existing_widgets = None

        # Connect the tree items
        self.treeWidget.itemChanged.connect(self.on_item_checked)

        # Get all existing items
        self.existing_widgets = self.get_existing_widgets()
        # Populate the tree widget
        self.populate_tree()
        # Restore the selection from Setting
        self.select_from_setting()
        # Open selected widgets
        self.show_widgets()


    def on_item_checked(self, item, column):
        if column != 0:
            return

        type_widget = item.data(0, Qt.UserRole)
        if type_widget is None:
            return

        # On crée l'entrée avec le nom et la classe
        entry = (item.text(0), type_widget)

        if item.checkState(0) == Qt.Checked:
            self.selected_widgets.append(entry)
        else:
            # On retire une occurrence si elle existe
            if entry in self.selected_widgets:
                self.selected_widgets.remove(entry)

        print("Selected for autoshow:", len(self.selected_widgets))


    def populate_tree(self):
        if not self.existing_widgets:
            self.error("No widget found on the Orange window.")
            return

        self.treeWidget.clear()
        # Optionnel : définir une seule colonne ou plusieurs
        self.treeWidget.setHeaderLabels(["Widgets", "Types"])

        # existing_widgets est ton dict { 'Data': [('File', <obj>), ...], ... }
        for category, widgets in self.existing_widgets.items():
            # Parent (Catégorie) - On l'étale sur les deux colonnes (ou pas)
            parent_item = QTreeWidgetItem([category])
            # --- ON REMET LE GRAS ICI ---
            font = parent_item.font(0)
            font.setBold(True)
            parent_item.setFont(0, font)

            self.treeWidget.addTopLevelItem(parent_item)

            for title, name, widget_obj in widgets:
                # Création de l'item avec 2 colonnes : [Titre, NomTechnique]
                child_item = QTreeWidgetItem([title, name])

                # --- Style de la colonne 0 (Le Titre) ---
                child_item.setFlags(child_item.flags() | Qt.ItemIsUserCheckable)
                child_item.setCheckState(0, Qt.Unchecked)
                child_item.setData(0, Qt.UserRole, type(widget_obj))

                # --- Style de la colonne 1 (Le Nom Technique) ---
                child_item.setFont(1, italic_font)
                child_item.setForeground(1, gray_color)

                parent_item.addChild(child_item)

        self.treeWidget.expandAll()


    def select_from_setting(self):
        self.warning("")
        if not self.selected_widgets:
            return

        tree = self.treeWidget
        tree.blockSignals(True)  # block signals so on_item_checked is NOT called
        try:
            valid_widgets = []

            for title, widget_type in self.selected_widgets:
                found = False

                # Iterate over top-level items (categories)
                for i in range(tree.topLevelItemCount()):
                    parent = tree.topLevelItem(i)

                    # Iterate over children (actual widgets)
                    for j in range(parent.childCount()):
                        child = parent.child(j)

                        # Match title + widget type
                        if (
                                child.text(0) == title
                                and child.data(0, Qt.UserRole) == widget_type
                        ):
                            # Only check if not already checked
                            if child.checkState(0) != Qt.Checked:
                                child.setCheckState(0, Qt.Checked)
                            found = True
                            break  # stop at first match

                    if found:
                        break

                # Keep only widgets that exist in the tree
                if found:
                    valid_widgets.append((title, widget_type))

            # Replace selected_widgets with valid ones
            if len(self.selected_widgets) != len(valid_widgets):
                diff = abs(len(self.selected_widgets) - len(valid_widgets))
                self.warning(f"{diff} widget(s) missing from the initial configuration.")
            self.selected_widgets = valid_widgets

        finally:
            tree.blockSignals(False)  # re-enable signals


    def show_widgets(self):
        for item in self.selected_widgets:
            title = item[0]
            widget_type = item[1]  # C'est le string "QPushButton" etc.

            # On parcourt chaque liste de widgets dans le dictionnaire
            for category_list in self.existing_widgets.values():
                for w in category_list:
                    # w[0] = title, w[2] = obj_widget
                    # On compare le titre ET le nom de la classe de l'objet
                    if w[0] == title and type(w[2]) == widget_type:
                        # On vérifie si la méthode reshow existe pour éviter un crash
                        if hasattr(w[2], 'reshow'):
                            w[2].reshow()


    def get_existing_widgets(self):
        # Get the MainWindow
        app = QApplication.instance()
        main_windows = []
        for item in app.topLevelWidgets():
            if isinstance(item, CanvasMainWindow):
                main_windows.append(item)
        if len(main_windows) == 0:
            self.error("Could not detect the main Orange window !")
            return
        elif len(main_windows) > 1:
            # TODO: Find main window with memory address of self
            self.error("Several main Orange windows detected ! Not implemented yet !")
            return

        scheme_edit_widget = main_windows[0].current_document()
        dict_of_items = scheme_edit_widget._SchemeEditWidget__widgetManager._OWWidgetManager__item_for_node

        existing_widgets = defaultdict(list)
        for key, value in dict_of_items.items():
            category = key.description.category
            title = key.title
            name = key.description.name
            obj_widget = value.widget
            # Plus besoin de vérifier si la catégorie existe !
            existing_widgets[category].append((title, name, obj_widget))
        return existing_widgets

if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWAutoShow()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()
